"""Types for acoupi ORM models."""

from datetime import datetime
from typing import List, NamedTuple, Optional
from uuid import UUID

from pony.orm import core

__all__ = [
    "BaseModels",
    "Deployment",
    "PredictedTag",
    "Recording",
    "Detection",
    "ModelOutput",
]


class Deployment(core.EntityMeta):
    """Deployment ORM model."""

    id: UUID
    """Unique ID of the deployment."""

    name: str
    """Deployment name"""

    started_on: datetime
    """Datetime when the deployment started. Should be unique."""

    ended_on: Optional[datetime]

    latitude: Optional[float]
    """Latitude of the deployment site. Can be None if unknown."""

    longitude: Optional[float]
    """Longitude of the deployment site. Can be None if unknown."""


class Recording(core.EntityMeta):
    """Recording ORM model."""

    id: UUID
    """Unique ID of the recording"""

    path: Optional[str]
    """Path to the recording file. Can be None if the recording is not stored locally"""

    datetime: datetime
    """Datetime of the recording. Should be unique"""

    duration_s: float
    """Duration of the recording in seconds"""

    samplerate_hz: int
    """Samplerate of the recording in Hz"""

    audio_channels: int
    """Number of audio_channels in the recording"""

    deployment: Deployment
    """Deployment that the recording belongs to"""

    model_outputs: List["ModelOutput"]
    """Model outputs that belong to the recording"""


class PredictedTag(core.EntityMeta):
    """Predicted tag ORM model."""

    id: int
    """Unique ID of the predicted tag"""

    key: str
    """Key of the predicted tag"""

    value: str
    """Value of the predicted tag"""

    confidence_score: float
    """Score of the predicted tag"""

    detection: Optional["Detection"]
    """Detection that the predicted tag belongs to"""

    model_output: Optional["ModelOutput"]
    """Model output that the predicted tag belongs to"""


class Detection(core.EntityMeta):
    """Detection ORM model."""

    id: UUID
    """Unique ID of the detection."""

    start_time_s: Optional[float]
    """Detection start time in seconds."""

    end_time_s: Optional[float]
    """Detection end time in seconds."""

    low_freq_hz: Optional[float]
    """Detection low frequency in Hz."""

    high_freq_hz: Optional[float]
    """Detection high frequency in Hz."""

    detection_score: float
    """Score of the detection."""

    tags: List[PredictedTag]
    """Predicted tags of the detection."""

    model_output: "ModelOutput"
    """Model output that the detection belongs to."""


class ModelOutput(core.EntityMeta):
    """Model output ORM model."""

    id: UUID
    """Unique ID of the model output"""

    model_name: str
    """Name of the model"""

    recording: Recording
    """Recording that the model output belongs to"""

    tags: List[PredictedTag]
    """Predicted tags of the model output at the file level"""

    detections: List[Detection]
    """Detections of the model output"""

    created_on: datetime
    """Datetime when the model output was created"""


class BaseModels(NamedTuple):
    """Container for models."""

    Deployment: Deployment
    Recording: Recording
    PredictedTag: PredictedTag
    Detection: Detection
    ModelOutput: ModelOutput



"""Test acoupi recording filters."""

from acoupi import components, data


def test_focus_species_filter_rejects_recordings_without_detections(
    recording: data.Recording,
):
    """Test focus species recording filter rejects recordings without detections."""
    filter = components.DetectionTags(
        tags=[data.Tag(key="species", value="test")],
        saving_threshold=0.5,
    )
    assert not filter.should_save_recording(recording, [])


def test_focus_species_filter_rejects_if_no_target_species_found(
    recording: data.Recording,
):
    """Test focus species recording filter rejects if no target species found."""
    filter = components.DetectionTags(
        tags=[data.Tag(key="species", value="test")],
        saving_threshold=0.5,
    )
    model_output = data.ModelOutput(
        name_model="test",
        recording=recording,
        detections=[
            data.Detection(
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test2"),
                        confidence_score=0.6,
                    ),
                ],
            ),
        ],
    )
    assert not filter.should_save_recording(recording, [model_output])


def test_focus_species_filter_rejects_low_confidence_detections(
    recording: data.Recording,
):
    """Test focus species recording filter rejects low confidence detections."""
    filter = components.DetectionTags(
        tags=[data.Tag(key="species", value="test")],
        saving_threshold=0.5,
    )
    model_output = data.ModelOutput(
        name_model="test",
        recording=recording,
        detections=[
            data.Detection(
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test2"),
                        confidence_score=0.4,
                    ),
                ],
            ),
        ],
    )
    assert not filter.should_save_recording(recording, [model_output])


def test_focus_species_filter_keeps_high_confidence_detections(
    recording: data.Recording,
):
    """Test focus species recording filter keeps high confidence detections."""
    filter = components.DetectionTags(
        tags=[data.Tag(key="species", value="test")],
        saving_threshold=0.5,
    )
    model_output = data.ModelOutput(
        name_model="test",
        recording=recording,
        detections=[
            data.Detection(
                detection_score=0.6,
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test"),
                        confidence_score=0.6,
                    ),
                ],
            ),
        ],
    )

    assert filter.should_save_recording(recording, [model_output])


def test_focus_species_filter_rejects_even_with_confident_non_target(
    recording: data.Recording,
):
    """Test focus species recording filter rejects correctly.

    Even in the case with confident non-target detections.
    """
    filter = components.DetectionTags(
        tags=[data.Tag(key="species", value="test")],
        saving_threshold=0.5,
    )
    model_output = data.ModelOutput(
        name_model="test",
        recording=recording,
        detections=[
            data.Detection(
                detection_score=0.6,
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test2"),
                        confidence_score=0.6,
                    ),
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test"),
                        confidence_score=0.4,
                    ),
                ],
            ),
        ],
    )
    assert not filter.should_save_recording(recording, [model_output])


def test_focus_species_filter_keeps_with_at_least_one_target_species(
    recording: data.Recording,
):
    """Test focus species recording filter keeps correctly.

    It should keep recordings with at least one target species detection
    with a score above the provided threshold.
    """
    filter = components.DetectionTags(
        tags=[
            data.Tag(key="species", value="test"),
            data.Tag(key="species", value="test2"),
        ],
        saving_threshold=0.5,
    )
    model_output = data.ModelOutput(
        name_model="test",
        recording=recording,
        detections=[
            data.Detection(
                tags=[
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test2"),
                        confidence_score=0.6,
                    ),
                    data.PredictedTag(
                        tag=data.Tag(key="species", value="test"),
                        confidence_score=0.4,
                    ),
                ],
            ),
        ],
    )
    assert filter.should_save_recording(recording, [model_output])



"""Detection Task.

This module contains the function to generate a detection task.
The detection task is a function that takes a recording as input and
runs the detection process on the recording. The detection process
contains the following steps:

1. Check if the recording should be processed by the model.
2. Run the model on the recording.
3. Clean the outputs of the model based on the output cleaners.
4. Store the cleaned outputs of the model in the store.
5. Create messages to be sent using the Messenger.
"""

import logging
from typing import Callable, List, Optional

from acoupi import data
from acoupi.components import types

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def generate_detection_task(
    store: types.Store,
    model: types.Model,
    message_store: types.MessageStore,
    logger: logging.Logger = logger,
    output_cleaners: Optional[List[types.ModelOutputCleaner]] = None,
    processing_filters: Optional[List[types.ProcessingFilter]] = None,
    message_factories: Optional[List[types.MessageBuilder]] = None,
) -> Callable[[data.Recording], None]:
    """Generate a detection task.

    Parameters
    ----------
    store : types.Store
        The store to store the model output.
    model : types.Model
        The model to run on the recording.
    message_store : types.MessageStore
        The message store to store the messages.
    logger : logging.Logger, optional
        The logger to log messages, by default logger.
    output_cleaners : Optional[List[types.ModelOutputCleaner]], optional
        The output cleaners to clean the model output, by default None.
    processing_filters : Optional[List[types.ProcessingFilter]], optional
        The processing filters to check if the recording should be processed,
        by default None.
    message_factories : Optional[List[types.MessageBuilder]], optional
        The message factories to create messages, by default None.

    Note
    -----
    The detection task calls the following methods:

    1. **filter.should_process_recording(recording)** -> bool
        - Check if the recording should be processed by the model.
        - See
        [components.processing_filters][acoupi.components.processing_filters]
        for implementations of
        [types.ProcessingFilter][acoupi.components.types.ProcessingFilter].
    2. **model.run(recording)** -> data.ModelOutput
        - Run the model on the recording and return the output.
        - See [types.Model][acoupi.components.types.Model].
    3. **cleaner.clean(model_output)** -> data.ModelOutput
        - Clean the outputs of the model based on the output cleaners.
        - See [components.output_cleaners][acoupi.components.output_cleaners]
        for implementations of
        [types.ModelOutputCleaner][acoupi.components.types.ModelOutputCleaner].
    4. **store.store_model_output(model_output)** -> None
        - Store the cleaned outputs of the model in the store.
        - See [components.stores][acoupi.components.stores] for implementation
        of [types.Store][acoupi.components.types.Store].
    5. **message_factory.build_message(model_output)** -> data.Message
        - Create messages to be sent using the Messenger.
        - See
        [components.message_factories][acoupi.components.message_factories] for
        implementations of
        [types.MessageBuilder][acoupi.components.types.MessageBuilder].
    6. **message_store.store_message(message)** -> None
        - Store the message in the message store.
        - See [components.message_stores][acoupi.components.message_stores] for
        implementation of [types.Store][acoupi.components.types.Store].
    """

    def detection_task(recording: data.Recording) -> None:
        """Run the detection process on a recording."""
        logger.info("Starting detection process on recording %s", recording)

        # Check if recording should be processed
        if not all(
            filter.should_process_recording(recording)
            for filter in processing_filters or []
        ):
            # If recording should not be processed exit process
            logger.info("Recording should not be processed, skipping.")
            return

        # Detect events in recordings
        logger.info(f"Running model on recording: {recording.path}")
        model_output = model.run(recording)

        # Clean model output
        for cleaner in output_cleaners or []:
            model_output = cleaner.clean(model_output)

        # Store detections
        logger.info("Storing model output.")
        store.store_model_output(model_output)

        # Create messages
        for message_factory in message_factories or []:
            message = message_factory.build_message(model_output)

            if message is not None:
                logger.info("Storing message.")
                message_store.store_message(message)

    return detection_task



"""Test suite for the CLI task commands."""

from pathlib import Path

import pytest
from click.testing import CliRunner

from acoupi import system
from acoupi.cli import acoupi


@pytest.fixture
def setup_program(settings: system.Settings):
    system.setup_program(settings, "acoupi.programs.test", prompt=False)


def test_can_get_the_full_list_of_test_program_tasks(
    setup_program,
    settings: system.Settings,
):
    runner = CliRunner()
    result = runner.invoke(
        acoupi,
        ["task", "list"],
        obj={"settings": settings},
    )
    assert result.exit_code == 0
    assert "test_task" in result.output


def test_can_save_the_task_profile_to_a_file(
    setup_program,
    settings: system.Settings,
    tmp_path: Path,
):
    runner = CliRunner()
    profile_path = tmp_path / "profile.prof"
    result = runner.invoke(
        acoupi,
        ["task", "profile", "test_task", "--output", str(profile_path)],
        obj={"settings": settings},
    )
    assert result.exit_code == 0
    assert profile_path.exists()



"""Test running the model of a recorded audio file."""

import datetime
from pathlib import Path

from acoupi import data
from acoupi.components.model_template import TestModel


def test_model(deployment: data.Deployment):
    """Test running the model."""
    recording = data.Recording(
        path=Path("tests/data/test_ultrasonic.wav"),
        duration=1,
        samplerate=16000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )
    model = TestModel()
    results = model.run(recording)
    assert isinstance(results, data.ModelOutput)



"""Jinja2 environment for rendering templates."""

from jinja2 import Environment, PackageLoader, select_autoescape

__all__ = [
    "render_template",
]


env = Environment(
    loader=PackageLoader("acoupi"),
    autoescape=select_autoescape(),
)


def render_template(template_name, **kwargs):
    """Render a template with the given name and keyword arguments."""
    template = env.get_template(template_name)
    return template.render(**kwargs)



"""Acoupi Workers CLI Commands.

This module provides command-line interface (CLI) commands for manually
managing Acoupi workers.

While _acoupi_ primarily relies on systemd for robust worker management in
deployments (initiated via `acoupi deployment start`), these commands offer
direct control over worker processes for testing and debugging purposes.

With these commands, you can:

- **Start Workers:** Manually launch worker processes to handle task execution.
- **Restart Workers:**  Restart workers to apply configuration
    changes or troubleshoot issues.
- **Stop Workers:**  Safely shut down worker processes.

This manual control can be valuable in development and testing scenarios where
interacting directly with the workers is necessary.
"""

import click

from acoupi import system
from acoupi.cli.base import acoupi

__all__ = [
    "workers",
]


@acoupi.group()
@click.pass_context
def workers(ctx):
    """Manage acoupi workers.

    Provides subcommands to manually control the Celery workers responsible for
    executing program tasks.
    """
    settings = ctx.obj["settings"]
    if not system.is_configured(settings):
        raise click.UsageError(
            "Acoupi is not setup. Run `acoupi setup` first."
        )


@workers.command()
@click.pass_context
def start(ctx):
    """Start the Celery workers.

    Manually starts the worker processes that handle the execution of program
    tasks.
    """
    settings: system.Settings = ctx.obj["settings"]
    system.start_workers(settings)


@workers.command()
@click.pass_context
def restart(ctx):
    """Restart the Celery workers.

    Restarts the worker processes, which can be useful for applying
    configuration changes or resolving issues.
    """
    settings: system.Settings = ctx.obj["settings"]
    system.restart_workers(settings)


@workers.command()
@click.pass_context
def stop(ctx):
    """Stop the Celery workers.

    Manually stops the worker processes, gracefully shutting down the task
    execution environment.
    """
    settings: system.Settings = ctx.obj["settings"]
    system.stop_workers(settings)


@workers.command()
@click.pass_context
def purge(ctx):
    """Purge all pending tasks from the Celery queue."""
    settings: system.Settings = ctx.obj["settings"]
    system.purge_queues(settings)



"""Basic Program template module.

This module provides a base class (`BasicProgram`) for creating basic Acoupi
programs. It offers essential features for audio recording, metadata storage,
and file management.

The `BasicProgram` class defines the following components:

- **Audio Recorder:**  Records audio clips according to the program's
  configuration.
- **Store:** Provides an interface for storing and retrieving metadata
  associated with the program and its recordings.

Using these components, `BasicProgram` creates and manages the following
tasks:

- **Audio Recording:** Records audio at regular intervals, configurable
  through the `audio` settings in the `BasicProgramConfiguration` schema.
- **File Management:**  Periodically performs file management operations,
  such as moving recordings from temporary to permanent storage.

Configuration:

The program's configuration is defined by the `BasicProgramConfiguration`
schema, which includes settings for:

- Timezone
- Microphone configuration
- Audio recording settings (duration, interval, schedule)
- Data storage locations (temporary directory, audio directory,
  metadata database)

Usage:

To create a basic Acoupi program, define a new class that inherits from
`BasicProgram`.

Customization:

You can customise the program's behavior by overriding the following methods:

- `get_recording_conditions`:  Modify the conditions that trigger audio
  recording.
- `get_recording_filters`:  Add filters to determine which recordings to
  save permanently.
- `get_recording_callbacks`: Define actions to perform after a recording
  is made.

"""

import datetime
import zoneinfo
from pathlib import Path
from typing import Annotated, Callable, Optional, TypeVar

from pydantic import BaseModel, Field
from pydantic_extra_types.timezone_name import TimeZoneName

from acoupi import components, data, tasks
from acoupi.components import types
from acoupi.components.audio_recorder import MicrophoneConfig
from acoupi.programs.core import (
    DEFAULT_WORKER_CONFIG,
    AcoupiProgram,
    NoUserPrompt,
)
from acoupi.system.files import get_temp_dir

__all__ = []


class AudioConfiguration(BaseModel):
    """Audio configuration schema."""

    duration: int = 3
    """Duration of each audio recording in seconds."""

    interval: int = 10
    """Interval between each audio recording in seconds."""

    chunksize: Annotated[int, NoUserPrompt] = 8192
    """Chunksize of audio recording."""

    schedule_start: datetime.time = Field(
        default=datetime.time(hour=6, minute=0, second=0),
    )
    """Start time for recording schedule."""

    schedule_end: datetime.time = Field(
        default=datetime.time(hour=22, minute=30, second=0),
    )
    """End time for recording schedule."""


class PathsConfiguration(BaseModel):
    """Data configuration schema."""

    tmp_audio: Path = Field(default_factory=get_temp_dir)
    """Temporary directory for storing audio files."""

    recordings: Path = Field(
        default_factory=lambda: Path.home() / "storages" / "recordings",
    )
    """Directory for storing audio files permanently."""

    db_metadata: Path = Field(
        default_factory=lambda: Path.home() / "storages" / "metadata.db",
    )
    """Path to the metadata database."""


class BasicProgramConfiguration(BaseModel):
    """Configuration schema for a basic program."""

    timezone: TimeZoneName = Field(default=TimeZoneName("Europe/London"))
    """Time zone where the device will be deployed."""

    microphone: MicrophoneConfig
    """Microphone configuration."""

    recording: AudioConfiguration = Field(default_factory=AudioConfiguration)
    """Audio configuration."""

    paths: PathsConfiguration = Field(default_factory=PathsConfiguration)
    """Data configuration."""


ProgramConfig = TypeVar("ProgramConfig", bound=BasicProgramConfiguration)


class BasicProgram(AcoupiProgram[ProgramConfig]):
    """Basic Acoupi Program.

    This class provides a base for creating basic Acoupi programs. It offers
    essential features for audio recording, metadata storage, and file
    management.

    Components:

    - **Audio Recorder:** Records audio clips according to the program's
      configuration.
    - **Store:** Provides an interface for storing and retrieving metadata
      associated with the program and its recordings.

    Tasks:

    Using the components above, this class creates and manages the following
    tasks:

    - **Audio Recording:** Records audio at regular intervals, configurable
      through the `audio` settings in the `BasicProgramConfiguration` schema.
    - **File Management:** Periodically performs file management operations,
      such as moving recordings from temporary to permanent storage.

    Customization:

    customise the program's behavior by overriding these methods:

    - `get_recording_conditions`: Define the specific conditions that must be
        met for audio recording to continue when the recording task is
        triggered by the scheduler.
    - `get_recording_filters`:  Add filters to determine which recordings to
      save.
    - `get_recording_callbacks`: Define actions to perform after a recording
      is made.

    Examples
    --------
    ```python
    import datetime
    from acoupi import components, data
    from acoupi.programs.templates import (
        BasicProgram,
        BasicProgramConfiguration,
    )


    class Config(BasicProgramConfiguration):
        pass


    class Program(BasicProgram):
        configuration_schema = Config

        def get_recording_conditions(self, config: Config):
            # Get the default recording conditions
            conditions = super().get_recording_conditions(
                config
            )
            return [
                components.IsInInterval(
                    data.TimeInterval(
                        start=datetime.time(hour=3),
                        end=datetime.time(hour=6),
                    )
                ),
                *conditions,
            ]
    ```
    """

    worker_config = DEFAULT_WORKER_CONFIG

    recorder: types.AudioRecorder

    store: types.Store

    def setup(self, config: ProgramConfig) -> None:
        """Set up the basic program.

        This method initialises the program's components (audio recorder,
        store, and file manager), registers the recording and file management
        tasks, and performs necessary setup operations.
        """
        self.validate_dirs(config)
        self.recorder = self.configure_recorder(config)
        self.store = self.configure_store(config)
        self.register_recording_task(config)
        self.register_file_management_task(config)
        super().setup(config)

    def on_start(self, deployment: data.Deployment):
        """Handle program start event.

        This method is called when the program starts and stores the
        deployment information in the metadata store.
        """
        self.store.store_deployment(deployment)
        super().on_start(deployment)

    def on_end(self, deployment: data.Deployment) -> None:
        """Handle program end event.

        This method is called when the program ends. It updates the
        deployment information in the metadata store, and ensure that
        remaining tasks are completed before the program is stopped.

        Tasks to check are:
        - file_management_task (if implemented). Check if there are remaining
        files in the temporary directory and move them to the correct directory.
        """
        self.store.update_deployment(deployment)
        super().on_end(deployment)

        tmp_audio_path = self.config.paths.tmp_audio
        tmp_files = list(tmp_audio_path.glob("*"))

        if len(tmp_files) > 0:
            print(
                f"Running file_management_task to manage {len(tmp_files)}"
                " remaining files in the temporary directory."
            )
            self.tasks["file_management_task"].apply()

    def check(self, config: ProgramConfig):
        """Check the program's components.

        This method performs checks on the program's components to ensure
        they are functioning correctly. Currently, it only checks the PyAudio
        recorder if it is being used.
        """
        recorder_check = getattr(self.recorder, "check", None)
        if callable(recorder_check):
            recorder_check()

        super().check(config)

    def configure_recorder(
        self,
        config: ProgramConfig,
    ) -> types.AudioRecorder:
        """Configure the audio recorder.

        This method creates and configures an instance of the `PyAudioRecorder`
        based on the provided configuration.

        Returns
        -------
        types.AudioRecorder
            The configured audio recorder instance.
        """
        microphone = config.microphone
        return components.PyAudioRecorder(
            duration=config.recording.duration,
            samplerate=microphone.samplerate,
            audio_channels=microphone.audio_channels,
            device_name=microphone.device_name,
            chunksize=config.recording.chunksize,
            audio_dir=config.paths.tmp_audio,
        )

    def configure_store(
        self,
        config: ProgramConfig,
    ) -> types.Store:
        """Configure the metadata store.

        This method creates and configures an instance of the `SqliteStore`
        based on the provided configuration.

        Returns
        -------
        types.Store
            The configured metadata store instance.
        """
        return components.SqliteStore(config.paths.db_metadata)

    def get_file_managers(
        self,
        config: ProgramConfig,
    ) -> list[types.RecordingSavingManager]:
        """Get the file managers.

        This method defines how audio recordings should be saved and managed.
        It returns a list of file managers that are responsible for determining
        the final storage location of each recording.

        When a recording is marked for saving, the program iterates through the
        list of file managers in order. Each manager can either:

        - Return a path where the recording should be saved.
        - Return `None` to indicate that it cannot handle the recording,
        allowing the next manager in the list to be used.

        By default, this method returns a list containing a single
        `DateFileManager`, which stores recordings under `YYYY/MM/DD/` and
        uses a filename template that includes the recording date, time, and
        ID.

        Override this method if you want to provide a different saving manager
        or a customised `DateFileManager(filename_template=...)`.

        Returns
        -------
        list[types.RecordingSavingManager]
            A list of file manager instances.
        """
        return [components.DateFileManager(config.paths.recordings)]

    def get_recording_conditions(
        self,
        config: ProgramConfig,
    ) -> list[types.RecordingCondition]:
        """Get the recording conditions.

        This method defines the conditions under which audio recording should
        be performed. By default, it uses the schedule defined in the
        configuration.

        Returns
        -------
        types.RecordingCondition
            A recording condition.
        """
        timezone = zoneinfo.ZoneInfo(config.timezone)
        return [
            components.IsInIntervals(
                intervals=[
                    data.TimeInterval(
                        start=config.recording.schedule_start,
                        end=datetime.datetime.strptime(
                            "23:59:59", "%H:%M:%S"
                        ).time(),
                    ),
                    data.TimeInterval(
                        start=datetime.datetime.strptime(
                            "00:00:00", "%H:%M:%S"
                        ).time(),
                        end=config.recording.schedule_end,
                    ),
                ],
                timezone=timezone,
            )
        ]

    def get_recording_filters(
        self,
        config: ProgramConfig,
    ) -> list[types.RecordingSavingFilter]:
        """Get the recording saving filters.

        This method defines filters that determine which recordings should be
        saved permanently. By default, it returns an empty list, meaning all
        recordings are saved.

        Returns
        -------
        list[types.RecordingSavingFilter]
            A list of recording saving filters.
        """
        return []

    def get_recording_callbacks(
        self,
        config: ProgramConfig,
    ) -> list[Callable[[Optional[data.Recording]], None]]:
        """Get the recording callbacks.

        This method defines callbacks to be executed after a recording is
        completed. By default, it returns an empty list.

        Returns
        -------
        list[Callable[[Optional[data.Recording]], None]]
            A list of recording callbacks.
        """
        return []

    def get_required_models(self, config: ProgramConfig) -> list[str]:
        """Get the required models for a recording to be considered ready.

        This method specifies which bioacoustic models must process a recording
        before it is considered "ready" to be moved from temporary storage.

        By default, no models are required, meaning recordings are immediately
        considered ready. However, you can override this method to define
        specific models that must process the recordings based on the program's
        configuration.

        Returns
        -------
        list[str]
            A list of model names that are required to process a recording
            before it is considered ready.
        """
        return []

    def create_recording_task(
        self,
        config: ProgramConfig,
    ) -> Callable[[], Optional[data.Recording]]:
        """Create the recording task.

        This method creates the task responsible for recording audio.

        Returns
        -------
        Callable[[], Optional[data.Recording]]
            The recording task.
        """
        return tasks.generate_recording_task(
            recorder=self.recorder,
            store=self.store,
            logger=self.logger.getChild("recording"),
            recording_conditions=self.get_recording_conditions(config),
        )

    def create_file_management_task(
        self,
        config: ProgramConfig,
    ) -> Callable[[], None]:
        """Create the file management task.

        This method creates the task responsible for managing audio files.

        Returns
        -------
        Callable[[], None]
            The file management task.
        """
        return tasks.generate_file_management_task(
            store=self.store,
            logger=self.logger.getChild("file_management"),
            file_managers=self.get_file_managers(config),
            file_filters=self.get_recording_filters(config),
            required_models=self.get_required_models(config),
            tmp_path=config.paths.tmp_audio,
        )

    def register_recording_task(
        self,
        config: ProgramConfig,
    ) -> None:
        """Register the recording task.

        This method registers the recording task with the program's scheduler.
        """
        recording_task = self.create_recording_task(config)
        self.add_task(
            function=recording_task,
            schedule=datetime.timedelta(seconds=config.recording.interval),
            callbacks=self.get_recording_callbacks(config),
            queue="recording",
        )

    def register_file_management_task(
        self,
        config: ProgramConfig,
    ) -> None:
        """Register the file management task.

        This method registers the file management task with the program's
        scheduler.
        """
        file_management_task = self.create_file_management_task(config)
        self.add_task(
            function=file_management_task,
            schedule=datetime.timedelta(minutes=2),
            queue="celery",
        )

    def validate_dirs(self, config: ProgramConfig):
        """Validate the directories used by the program.

        This method ensures that the necessary directories for storing audio
        and metadata exist. If they don't, it creates them.
        """
        if not config.paths.tmp_audio.exists():
            config.paths.tmp_audio.mkdir(parents=True)

        if not config.paths.recordings.exists():
            config.paths.recordings.mkdir(parents=True)

        if not config.paths.db_metadata.parent.exists():
            config.paths.db_metadata.parent.mkdir(parents=True)



"""CLI for acoupi."""

from acoupi.cli.base import acoupi
from acoupi.cli.config import config
from acoupi.cli.deployment import deployment
from acoupi.cli.tasks import task
from acoupi.cli.workers import workers

__all__ = [
    "acoupi",
    "config",
    "deployment",
    "task",
    "workers",
]


if __name__ == "__main__":
    acoupi()



"""Module for detecting and identifying devices."""

import socket

__all__ = [
    "get_rpi_serial_number",
    "get_rpi_host_name",
    "is_rpi",
]


def get_rpi_serial_number() -> str:
    """Get the serial number of the Raspberry Pi.

    Returns
    -------
        The serial number of the Raspberry Pi as a string.
    """
    with open("/proc/cpuinfo", "r") as f:
        for line in f:
            if line[0:6] == "Serial":
                return line[10:26]
    raise RuntimeError("Could not find serial number of Raspberry Pi")


def get_rpi_host_name() -> str:
    """Get the hostname of the Raspberry Pi.

    Returns
    -------
        The hostname of the Raspberry Pi as a string.
    """
    return socket.gethostname()


def is_rpi() -> bool:
    """Check if the current device is a Raspberry Pi.

    Returns
    -------
        True if the current device is a Raspberry Pi, False otherwise.
    """
    try:
        get_rpi_serial_number()
        return True
    except FileNotFoundError:
        return False
    except RuntimeError:
        return False



"""Test Acoupi file managers."""

import datetime
from pathlib import Path
from unittest.mock import patch

import pytest

from acoupi import components, data
from acoupi.components import saving_managers


def test_save_recording_manager_fails_if_recording_has_no_path(
    deployment: data.Deployment,
    tmp_path: Path,
):
    manager = components.SaveRecordingManager(tmp_path)
    recording = data.Recording(
        path=None,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )

    with pytest.raises(ValueError):
        manager.save_recording(recording)


def test_save_recording_with_confident_tags(tmp_path: Path):
    audio_dir = tmp_path / "audio"
    tmp_dir = tmp_path / "tmp"
    tmp_dir.mkdir()
    recording_file = tmp_dir / "recording.wav"
    recording_file.touch()
    recording_file.write_text("test content")
    recording = data.Recording(
        path=recording_file,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=data.Deployment(name="test"),
    )
    model_output = data.ModelOutput(
        name_model="test_model",
        recording=recording,
        tags=[
            data.PredictedTag(
                tag=data.Tag(key="test", value="value1"),
                confidence_score=0.6,
            ),
            data.PredictedTag(
                tag=data.Tag(key="test", value="value2"),
                confidence_score=0.3,
            ),
        ],
    )
    manager = components.SaveRecordingManager(
        audio_dir, detection_threshold=0.6, saving_threshold=0.5
    )

    new_path = manager.save_recording(recording, model_outputs=[model_output])

    assert new_path is not None
    assert new_path.parent.parent == audio_dir
    assert new_path.parent == audio_dir / "true_detections"


def test_save_recording_with_unconfident_tags(tmp_path: Path):
    audio_dir = tmp_path / "audio"
    tmp_dir = tmp_path / "tmp"
    tmp_dir.mkdir()
    recording_file = tmp_dir / "recording.wav"
    recording_file.touch()
    recording_file.write_text("test content")
    recording = data.Recording(
        path=recording_file,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=data.Deployment(name="test"),
    )
    model_output = data.ModelOutput(
        name_model="test_model",
        recording=recording,
        tags=[
            data.PredictedTag(
                tag=data.Tag(key="test", value="value1"),
                confidence_score=0.4,
            ),
            data.PredictedTag(
                tag=data.Tag(key="test", value="value2"),
                confidence_score=0.3,
            ),
        ],
    )
    manager = components.SaveRecordingManager(
        audio_dir,
        detection_threshold=0.6,
        saving_threshold=0.4,
    )

    new_path = manager.save_recording(recording, model_outputs=[model_output])

    assert new_path is not None
    assert new_path.parent.parent == audio_dir
    assert new_path.parent == audio_dir / "false_detections"


def test_date_file_manager_save_recording(
    tmp_path: Path,
    deployment: data.Deployment,
):
    """Test DateFileManager.save_recording."""
    # Arrange
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"

    # create a recording
    year = 2023
    month = 4
    day = 15
    hour = 18
    minute = 9
    second = 30
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime(year, month, day, hour, minute, second),
        deployment=deployment,
    )
    # make sure the recording file exists
    path.touch()

    # create a file manager
    file_manager = components.DateFileManager(directory)

    # Act
    file_path = file_manager.save_recording(recording)

    # Assert
    assert directory.exists()
    assert file_path == (
        directory / "2023" / "4" / "15" / f"20230415_180930_{recording.id}.wav"
    )


def test_date_file_manager_supports_custom_filename_template(
    tmp_path: Path,
    deployment: data.Deployment,
):
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"
    recording = data.Recording(
        path=path,
        duration=1.5,
        samplerate=48000,
        audio_channels=2,
        created_on=datetime.datetime(2023, 4, 15, 18, 9, 30),
        deployment=data.Deployment(
            name="My Site/1",
            latitude=51.5,
            longitude=-0.1,
            started_on=deployment.started_on,
        ),
    )
    path.touch()

    with patch(
        "acoupi.components.saving_managers.get_device_info",
        return_value=data.DeviceInfo(id=""),
    ):
        file_manager = components.DateFileManager(
            directory,
            filename_template=(
                "{deployment.name}_{device.id}_{recording.created_on:%H%M}.wav"
            ),
        )
        file_path = file_manager.save_recording(recording)

    assert file_path == (
        directory / "2023" / "4" / "15" / "My Site_1__1809.wav"
    )


def test_date_file_manager_uses_empty_device_id_when_unavailable(
    tmp_path: Path,
):
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=48000,
        created_on=datetime.datetime(2023, 4, 15, 18, 9, 30),
        deployment=data.Deployment(name="site-a"),
    )
    path.touch()

    with patch(
        "acoupi.devices.get_device_id",
        side_effect=RuntimeError("device unavailable"),
    ):
        from acoupi.devices import get_device_info

        get_device_info.cache_clear()
        file_path = components.DateFileManager(
            directory,
            filename_template="{device.id}_{recording.created_on:%H%M}.wav",
        ).save_recording(recording)

    assert file_path == directory / "2023" / "4" / "15" / "_1809.wav"


def test_date_file_manager_fails_with_invalid_filename_template(
    tmp_path: Path,
    deployment: data.Deployment,
):
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime(2023, 4, 15, 18, 9, 30),
        deployment=deployment,
    )
    path.touch()

    file_manager = components.DateFileManager(
        directory,
        filename_template="{recording.hour:02d}.wav",
    )

    with pytest.raises(ValueError, match="Invalid filename template"):
        file_manager.save_recording(recording)


def test_date_file_manager_fails_if_recording_has_no_path(
    tmp_path: Path, deployment: data.Deployment
):
    """Test DateFileManager.save_recording fails if recording has no path."""
    # Arrange
    directory = tmp_path / "recordings"

    # create a recording
    year = 2023
    month = 4
    day = 15
    hour = 18
    minute = 9
    second = 30
    recording = data.Recording(
        path=None,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime(year, month, day, hour, minute, second),
        deployment=deployment,
    )

    # create a file manager
    file_manager = components.DateFileManager(directory)

    # Act and Assert
    with pytest.raises(ValueError):
        file_manager.save_recording(recording)


def test_date_file_manager_fails_if_recording_file_does_not_exist(
    tmp_path: Path, deployment: data.Deployment
):
    """Test DateFileManager.save_recording fails if recording file does not
    exist.
    """
    # Arrange
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"

    # create a recording
    year = 2023
    month = 4
    day = 15
    hour = 18
    minute = 9
    second = 30
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime(year, month, day, hour, minute, second),
        deployment=deployment,
    )

    # make sure the recording file does not exist
    try:
        path.unlink()
    except FileNotFoundError:
        pass

    # create a file manager
    file_manager = components.DateFileManager(directory)

    # Act and Assert
    with pytest.raises(FileNotFoundError):
        file_manager.save_recording(recording)


def test_id_file_manager_save_recording(
    tmp_path: Path, deployment: data.Deployment
):
    """Test IDFileManager.save_recording."""
    # Arrange
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"

    # create a recording
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )
    # make sure the recording file exists
    path.touch()

    # create a file manager
    file_manager = components.IDFileManager(directory)

    # Act
    file_path = file_manager.save_recording(recording)

    # Assert
    assert directory.exists()
    assert file_path == directory / f"{recording.id}.wav"


def test_id_file_manager_fails_if_recording_has_no_path(
    tmp_path: Path, deployment: data.Deployment
):
    """Test IDFileManager.save_recording fails if recording has no path."""
    # Arrange
    directory = tmp_path / "recordings"

    # create a recording
    recording = data.Recording(
        path=None,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )

    # create a file manager
    file_manager = components.IDFileManager(directory)

    # Act and Assert
    with pytest.raises(ValueError):
        file_manager.save_recording(recording)


def test_id_file_manager_fails_if_recording_file_does_not_exist(
    tmp_path: Path, deployment: data.Deployment
):
    """Test IDFileManager.save_recording fails if recording file does not
    exist.
    """
    # Arrange
    path = tmp_path / "test.wav"
    directory = tmp_path / "recordings"

    # create a recording
    recording = data.Recording(
        path=path,
        duration=1,
        samplerate=8000,
        created_on=datetime.datetime.now(),
        deployment=deployment,
    )

    # make sure the recording file does not exist
    try:
        path.unlink()
    except FileNotFoundError:
        pass

    # create a file manager
    file_manager = components.IDFileManager(directory)

    # Act and Assert
    with pytest.raises(FileNotFoundError):
        file_manager.save_recording(recording)



"""Acoupi components."""

from acoupi.components.audio_recorder import MicrophoneConfig, PyAudioRecorder
from acoupi.components.message_factories import (
    DetectionThresholdMessageBuilder,
    FullModelOutputMessageBuilder,
)
from acoupi.components.message_stores.sqlite import SqliteMessageStore
from acoupi.components.messengers import (
    HTTPConfig,
    HTTPMessenger,
    MQTTConfig,
    MQTTMessenger,
)
from acoupi.components.output_cleaners import ThresholdDetectionCleaner
from acoupi.components.recording_conditions import IsInInterval, IsInIntervals
from acoupi.components.recording_schedulers import IntervalScheduler
from acoupi.components.saving_filters import (
    After_DawnDuskTimeInterval,
    Before_DawnDuskTimeInterval,
    DetectionTags,
    DetectionTagValue,
    FrequencySchedule,
    SaveIfInInterval,
    SavingThreshold,
)
from acoupi.components.saving_managers import (
    DateFileManager,
    IDFileManager,
    SaveRecordingManager,
)
from acoupi.components.stores.sqlite import SqliteStore
from acoupi.components.summariser import (
    StatisticsDetectionsSummariser,
    ThresholdsDetectionsSummariser,
)

__all__ = [
    "After_DawnDuskTimeInterval",
    "Before_DawnDuskTimeInterval",
    "DateFileManager",
    "DetectionTagValue",
    "DetectionTags",
    "DetectionThresholdMessageBuilder",
    "FrequencySchedule",
    "FullModelOutputMessageBuilder",
    "HTTPConfig",
    "HTTPMessenger",
    "IDFileManager",
    "IntervalScheduler",
    "IsInInterval",
    "IsInIntervals",
    "MQTTConfig",
    "MQTTMessenger",
    "MicrophoneConfig",
    "PyAudioRecorder",
    "SaveIfInInterval",
    "SaveRecordingManager",
    "SavingThreshold",
    "SqliteMessageStore",
    "SqliteStore",
    "StatisticsDetectionsSummariser",
    "ThresholdDetectionCleaner",
    "ThresholdsDetectionsSummariser",
]



"""Path constants for acoupi system."""

from pathlib import Path
from typing import List, Literal

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings, SettingsConfigDict

__all__ = ["Settings", "CeleryConfig"]


class Settings(BaseSettings):
    """Settings for acoupi system."""

    model_config = SettingsConfigDict(
        env_prefix="ACOUPI_",
    )

    home: Path = Path.home() / ".acoupi"
    app_name: str = "app"
    program_file: Path = home / (app_name + ".py")
    program_name_file: Path = home / "config" / "name"
    program_config_file: Path = home / "config" / "program.json"
    celery_config_file: Path = home / "config" / "celery.json"
    deployment_file: Path = home / "config" / "deployment.json"
    env_file: Path = home / "config" / "env"
    run_dir: Path = home / "run"
    log_dir: Path = home / "log"
    log_level: str = "INFO"
    start_script_path: Path = home / "bin" / "acoupi-workers-start.sh"
    stop_script_path: Path = home / "bin" / "acoupi-workers-stop.sh"
    restart_script_path: Path = home / "bin" / "acoupi-workers-restart.sh"
    beat_script_path: Path = home / "bin" / "acoupi-beat.sh"
    acoupi_service_file: str = "acoupi.service"
    acoupi_beat_service_file: str = "acoupi-beat.service"
    celery_pool_type: Literal[
        "threads",
        "prefork",
        "eventlet",
        "gevent",
        "solo",
        "processes",
    ] = "threads"


class CeleryConfig(BaseModel):
    """Configuration settings for Celery in Acoupi.

    This class defines the settings used to configure the Celery
    task queue specifically for the Acoupi application.
    """

    enable_utc: bool = True
    """Whether to enable UTC for Celery.

    It's generally recommended to keep this enabled for consistency across
    different Acoupi deployments.
    """

    timezone: str = "UTC"
    """The timezone to use for Celery."""

    broker_url: str = "pyamqp://guest@localhost//"
    """The URL of the message broker used by Celery.

    Acoupi uses RabbitMQ with the default guest user. You may need to
    update this if your RabbitMQ setup is different.
    """

    broker_connection_retry_on_startup: bool = True
    """Retry to establish the connection to the AMQP broker on startup.

    Automatically try to re-establish the connection to the AMQP broke
    if lost after the initial connection is made."""

    result_backend: str = "rpc://"
    """ The URL for storing task results. 

    'rpc://' indicates that results are sent back directly to the client.
    """

    result_persistent: bool = False
    """Whether to persist task results. 

    In Acoupi deployments, task results are not typically needed after the task
    has completed, as all essential data is stored in an independent database.
    This setting helps to avoid unnecessary storage overhead.
    """

    task_serializer: str = "pickle"
    result_serializer: str = "pickle"
    accept_content: List[str] = Field(default_factory=lambda: ["pickle"])

    task_acks_late: bool = True
    """Whether to acknowledge tasks after they have been executed. 
    
    True means that tasks are acknowledged after they have been executed, 
    not right before.
    """

    worker_prefetch_multiplier: int = 1
    """The number of tasks a worker can prefetch.

    Setting this to 1 prevents tasks from being delayed due to other tasks in
    the queue. Celery defaults to prefetching tasks in batches, which can cause
    a fast task to wait for a slower one in the same batch.
    """

    task_soft_time_limit: int = 30
    """The soft time limit (in seconds) for task execution.

    If a task exceeds this limit, it will receive a warning.

    If you have tasks that are expected to run longer than this limit,
    you should increase this value or specify the time limit directly.
    """

    task_time_limit: int = 60
    """The hard time limit (in seconds) for task execution.

    If a task exceeds this limit, it will be terminated.

    If you have tasks that are expected to run longer than this limit,
    you should increase this value or specify the time limit directly.
    """



"""Acoupi Workers Module.

This module provides the functionality for managing and configuring
AcoupiWorkers, which are essential components of the Acoupi framework for
building smart acoustic sensors. AcoupiWorkers are responsible for executing
tasks related to acoustic sensing and analysis. Each Acoupi program relies on a
set of AcoupiWorkers, each with a unique name and configuration.

The AcoupiWorker is implemented as a Celery worker, which listens to designated
queues for task execution and coordination.

"""

from typing import List, Optional

from pydantic import BaseModel, Field, field_validator

__all__ = [
    "AcoupiWorker",
    "WorkerConfig",
    "DEFAULT_WORKER_CONFIG",
]


class AcoupiWorker(BaseModel):
    """AcoupiWorker Class.

    Represents an individual worker instance within the Acoupi framework. It is
    responsible for executing tasks related to acoustic sensing and analysis.
    Each worker is implemented as a Celery worker and listens to designated
    queues for task execution and coordination.
    """

    name: str
    """Name of the worker. Should be unique among different workers."""

    queues: List[str] = Field(default_factory=list)
    """Queues the worker should listen to. 

    If empty, the worker will listen to all queues."""

    concurrency: Optional[int] = None
    """Number of concurrent tasks the worker should run.

    If None, the worker will run as many tasks as possible.

    This setting should be set to 1 if the worker if at most
    one task can be run at a time. This is useful for tasks
    that are not thread safe or that require use of a resource
    that can only be used by one task at a time.
    """


class WorkerConfig(BaseModel):
    """WorkerConfig Class.

    A configuration class used to define the set of workers to be used for an
    Acoupi program. It allows you to specify the details of each worker, such
    as its name, concurrency level, and the queues it listens to.
    """

    workers: List[AcoupiWorker] = Field(default_factory=list)

    @field_validator("workers")
    def unique_worker_names(cls, workers: List[AcoupiWorker]):
        """Validate that all workers have unique names."""
        names = [worker.name for worker in workers]
        if len(set(names)) != len(names):
            raise ValueError("Worker names must be unique.")
        return workers


DEFAULT_WORKER_CONFIG = WorkerConfig(
    workers=[
        AcoupiWorker(
            name="recording",
            queues=["recording"],
            concurrency=1,
        ),
        AcoupiWorker(
            name="default",
            queues=["celery"],
        ),
    ],
)



"""Test suite for system deployment functions."""

import pytest

from acoupi import data
from acoupi.system import Settings, deployments, exceptions, programs


def test_start_deployment_saves_deployment_in_file(settings: Settings):
    """Test that create_deployment saves the deployment in the file."""
    deployment = deployments.start_deployment(settings, name="test")

    assert isinstance(deployment, data.Deployment)
    assert settings.deployment_file.exists()
    contents = settings.deployment_file.read_text()
    recovered = data.Deployment.model_validate_json(contents)
    assert recovered == deployment
    assert recovered.name == "test"
    assert recovered.started_on is not None
    assert recovered.ended_on is None


def test_start_deployment_with_location_info(settings: Settings):
    """Test that create_deployment saves the deployment in the file."""
    deployment = deployments.start_deployment(
        settings,
        name="test",
        latitude=1,
        longitude=2,
    )

    assert isinstance(deployment, data.Deployment)
    assert deployment.latitude == 1
    assert deployment.longitude == 2
    assert settings.deployment_file.exists()
    contents = settings.deployment_file.read_text()
    recovered = data.Deployment.model_validate_json(contents)
    assert recovered == deployment


def test_start_deployment_fails_if_already_deployed(settings: Settings):
    """Test that create_deployment fails if already deployed."""
    deployments.start_deployment(settings, name="test")
    with pytest.raises(exceptions.DeploymentError):
        deployments.start_deployment(settings, name="test2")


def test_end_deployment_saves_end_time(settings: Settings):
    """Test that end_deployment saves the end time."""
    deployments.start_deployment(settings, name="test")
    deployments.end_deployment(settings)

    assert settings.deployment_file.exists()
    contents = settings.deployment_file.read_text()
    deployment = data.Deployment.model_validate_json(contents)
    assert deployment.ended_on is not None


def test_start_deployment_after_ending_previous(settings: Settings):
    """Test that create_deployment works after ending the previous."""
    deployments.start_deployment(settings, name="test")
    deployments.end_deployment(settings)
    deployments.start_deployment(settings, name="test2")

    assert settings.deployment_file.exists()
    contents = settings.deployment_file.read_text()
    deployment = data.Deployment.model_validate_json(contents)
    assert deployment.name == "test2"
    assert deployment.started_on is not None
    assert deployment.ended_on is None


def test_end_deployment_fails_if_not_deployed(settings: Settings):
    """Test that end_deployment fails if not deployed."""
    with pytest.raises(exceptions.DeploymentError):
        deployments.end_deployment(settings)


def test_get_deployment_fails_if_hasnt_been_deployed(settings: Settings):
    """Test that get_deployment fails if the deployment hasn't been deployed."""
    with pytest.raises(exceptions.DeploymentError):
        deployments.get_current_deployment(settings)


def test_get_current_deployment(settings: Settings):
    """Test that get_deployment returns the current deployment."""
    deployments.start_deployment(
        settings, name="test", latitude=1, longitude=2
    )
    deployment = deployments.get_current_deployment(settings)
    assert isinstance(deployment, data.Deployment)
    assert deployment.name == "test"
    assert deployment.latitude == 1
    assert deployment.longitude == 2
    assert deployment.started_on is not None
    assert deployment.ended_on is None


def test_get_current_deployment_fails_if_ended(settings: Settings):
    """Test that get_deployment fails if the deployment has ended."""
    deployments.start_deployment(settings, name="test")
    deployments.end_deployment(settings)
    with pytest.raises(exceptions.DeploymentError):
        deployments.get_current_deployment(settings)



"""Detection Program template module.

This module provides a base class (`DetectionProgram`) for creating Acoupi
programs that perform audio detection tasks. It extends the `MessagingProgram`
class to add detection capabilities.

Key Components:

- **Detection Model:** An audio detection model that processes recordings to
  identify specific sounds or events.

Tasks:

- **Detection Task:** Runs the detection model on audio recordings and
  processes the results, including generating messages based on detected
events. Detections below a specified threshold are ignored.

Features:

- **Messaging Features:** Inherits all features from `MessagingProgram`,
  including:
    - **Messaging:** Sends messages and heartbeats to a configured messenger
      (HTTP or MQTT).
    - **Message Store:** A database for storing messages before they are sent.
- **Basic Program Features:** Inherits all features from `BasicProgram`,
  including:
    - **Audio Recording:** Records audio clips at regular intervals.
    - **Metadata Storage:** Stores metadata associated with recordings in a
      SQLite database.
    - **File Management:** Manages the storage of audio recordings, saving
      them to permanent storage and handling temporary files.

Usage:

To create an Acoupi program with audio detection capabilities:

1. Define a new class that inherits from `DetectionProgram`.
2. Define a configuration schema that inherits from
   `DetectionProgramConfiguration`.
3. Implement the `configure_model` method to configure your detection model.
4. Optionally override the following methods to customise the detection
   process:
    - `get_output_cleaners`:  To clean up the model's raw output.
    - `get_processing_filters`:  To filter recordings before processing.
    - `get_message_factories`:  To customise the messages generated based on
      detection results.
"""

import datetime
from abc import ABC, abstractmethod
from typing import Callable, List, TypeVar

from pydantic import BaseModel, Field

from acoupi import data
from acoupi.components import ThresholdDetectionCleaner, types
from acoupi.programs.core.workers import AcoupiWorker, WorkerConfig
from acoupi.programs.templates.messaging import (
    MessagingProgram,
    MessagingProgramConfiguration,
)
from acoupi.tasks import generate_detection_task

__all__ = [
    "DetectionProgram",
    "DetectionProgramConfiguration",
    "DetectionsConfiguration",
]

ModelConfig = TypeVar("ModelConfig", bound=BaseModel)


class DetectionsConfiguration(BaseModel):
    """Detection settings schema."""

    threshold: float = 0.2
    """Detections with a score below this threshold will be ignored."""


class DetectionProgramConfiguration(
    MessagingProgramConfiguration,
):
    """Detection Program Configuration schema.

    This schema extends the `MessagingProgramConfiguration` to include any
    additional settings required for detection programs.
    """

    detections: DetectionsConfiguration = Field(
        default_factory=DetectionsConfiguration
    )
    """Detection settings."""


C = TypeVar("C", bound=DetectionProgramConfiguration)


class DetectionProgram(MessagingProgram[C], ABC):
    """Detection Program.

    This abstract class extends the `MessagingProgram` to provide a foundation
    for detection programs. It includes functionality for configuring and
    running a detection model, processing model outputs, and generating
    messages based on detection results.

    Components:

    - **Detection Model:** An audio detection model that processes recordings
      to identify specific sounds or events.

    Tasks:

    - **Detection Task:** Runs the detection model on audio recordings and
      processes the results, including:
        - Cleaning the model output using `get_output_cleaners`. By default
        this includes a threshold cleaner that filters out detections below a
        specified threshold.
        - Filtering the processed output using `get_processing_filters`.
        - Generating messages based on the results using
          `get_message_factories`.

    Inherited Components:

    This class inherits the following components from `MessagingProgram`:

    - **Messenger:** A component responsible for sending messages and
      heartbeats via a configured communication protocol (HTTP or MQTT).
    - **Message Store:** A database for storing messages before they are sent.

    This class also inherits the following components from `BasicProgram`:

    - **Audio Recorder:** Records audio clips according to the program's
      configuration.
    - **File Manager:** Manages the storage of audio recordings, including
      saving them to permanent storage and handling temporary files.
    - **Store:** Provides an interface for storing and retrieving metadata
      associated with the program and its recordings.

    Inherited Tasks:

    This program inherits the following tasks from `MessagingProgram`:

    - **Heartbeat Task:** Periodically sends heartbeat messages to indicate
      that the program is running.
    - **Send Messages Task:** Periodically sends messages from the message
      store to the configured messenger.

    This program also inherits the following tasks from `BasicProgram`:

    - **Audio Recording:**  Records audio at regular intervals, configurable
      through the `audio` settings in the `BasicProgramConfiguration` schema.
    - **File Management:**  Periodically performs file management operations,
      such as moving recordings from temporary to permanent storage.

    Customization:

    You can customise the detection process by overriding the following
    methods:

    - `configure_model`:  **Required.** Implement this method to configure
      and return an instance of your detection model.
    - `get_output_cleaners`:  To clean up the model's raw output. Either
      override this method completely or call `super().get_output_cleaners()`
      to include the default threshold cleaner.
    - `get_processing_filters`:  To filter recordings before processing.
    - `get_message_factories`:  To customise the messages generated based on
      detection results.

    Examples
    --------
    ```python
    from acoupi.programs.templates import (
        DetectionProgram,
        DetectionProgramConfiguration,
    )
    from acoupi.components import types

    # This import should be replaced with your actual model import
    # This model does not exist in the acoupi package
    from acoupi.models import SimpleBirdModel


    class MyBirdDetectionConfiguration(
        DetectionProgramConfiguration
    ):
        # Add any configuration specific to bird detection
        pass


    class MyBirdDetectionProgram(
        DetectionProgram[MyBirdDetectionConfiguration]
    ):
        configuration_schema = MyBirdDetectionConfiguration

        def configure_model(
            self, config: MyBirdDetectionConfiguration
        ) -> types.Model:
            return (
                SimpleBirdModel()
            )  # Replace with your actual model
    ```
    """

    worker_config = WorkerConfig(
        workers=[
            AcoupiWorker(
                name="recording",
                queues=["recording"],
                concurrency=1,
            ),
            AcoupiWorker(
                name="detection",
                queues=["detection"],
                concurrency=1,
            ),
            AcoupiWorker(
                name="default",
                queues=["celery"],
            ),
        ]
    )

    model: types.Model
    """The configured detection model instance."""

    def register_recording_task(
        self,
        config: C,
    ) -> None:
        """Register the recording task.

        Overrides ``BasicProgram.register_recording_task`` to route detection
        callbacks to the dedicated ``"detection"`` worker (concurrency=1),
        preventing OOM from parallel model loads.

        Note: intentionally does not call ``super()`` — the base
        implementation would register the same recording task without the
        ``callback_queue`` argument, leaving callbacks unrouted.
        """
        recording_task = self.create_recording_task(config)
        self.add_task(
            function=recording_task,
            schedule=datetime.timedelta(seconds=config.recording.interval),
            callbacks=self.get_recording_callbacks(config),
            queue="recording",
            callback_queue="detection",
        )

    def setup(self, config: C) -> None:
        """Set up the Detection Program.

        This method initialises the detection model and performs any
        necessary setup.
        """
        self.model = self.configure_model(config)
        super().setup(config)

    def on_end(self, deployment: data.Deployment) -> None:
        """Handle program end event.

        This method is called when the program ends. It updates the
        deployment information in the metadata store, and ensure that
        remaining tasks are completed before the program is stopped.

        Tasks to check are:
        - file_management_task (if implemented). Check if there are remaining
        files in the temporary directory and move them to the correct directory.
        - detection_task (if implemented). Check if there are remaining files
        to be processed and process them.
        """
        super().on_end(deployment)

        tmp_audio_path = self.config.paths.tmp_audio
        remaining_files = list(tmp_audio_path.glob("*"))

        if len(remaining_files) > 0:
            print(
                f"Running detection_task on remaining {len(remaining_files)}."
            )
            recordings = self.store.get_recordings_by_path(remaining_files)
            for recording, _ in recordings:
                self.tasks["detection_task"].apply((recording,))
            print("Running final file_management_task.")
            self.tasks["file_management_task"].apply()

            if self.messenger is None:
                return
            print("Ending program. Running final messaging task.")
            self.tasks["send_messages_task"].apply()

    def check(self, config: C):
        """Check the program's components.

        This method performs checks on the program's components, including
        the detection model, to ensure they are functioning correctly.

        Parameters
        ----------
        config : C
            The program's configuration.
        """
        model_check = getattr(self.model, "check", None)
        if model_check and callable(model_check):
            model_check()

        super().check(config)

    @abstractmethod
    def configure_model(self, config: C) -> types.Model:
        """Configure the detection model.

        This method must be implemented by subclasses to configure and return
        an instance of the detection model.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        types.Model
            The configured detection model instance.
        """

    def get_output_cleaners(self, config: C) -> List[types.ModelOutputCleaner]:
        """Get the model output cleaners.

        This method can be overridden to define a list of output cleaners that
        will be applied to the model's raw output to clean it up or extract
        relevant information. By default, it includes a threshold cleaner that
        filters out detections below a specified threshold.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        List[types.ModelOutputCleaner]
            A list of model output cleaners.
        """
        if not config.detections.threshold:
            return []

        return [
            ThresholdDetectionCleaner(
                detection_threshold=config.detections.threshold
            )
        ]

    def get_processing_filters(
        self,
        config: C,
    ) -> List[types.ProcessingFilter]:
        """Get the processing filters.

        This method can be overridden to define a list of processing filters
        that will be applied to each recording *before* it is processed by the
        model. These filters determine whether a recording should be processed
        at all.

        This can be useful to avoid unnecessary model processing when it is not
        required by the context or based on simple heuristics on the recording
        content. Model processing can be computationally expensive, so it is
        beneficial to avoid it if possible.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        List[types.ProcessingFilter]
            A list of processing filters.
        """
        return []

    def get_message_factories(self, config: C) -> List[types.MessageBuilder]:
        """Get the message factories.

        This method can be overridden to define a list of message factories
        that will be used to generate messages based on the processed
        detection results.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        List[types.MessageBuilder]
            A list of message factories.
        """
        return []

    def create_detection_task(
        self,
        config: C,
    ) -> Callable[[data.Recording], None]:
        """Create the detection task.

        This method creates the task responsible for running the detection
        model and processing its output.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        Callable[[data.Recording], None]
            The detection task.

        Notes
        -----
        This method uses the `generate_detection_task` function to create the
        detection task. You can override this method to customise the task
        creation process.
        """
        return generate_detection_task(
            store=self.store,
            model=self.model,
            message_store=self.message_store,
            logger=self.logger.getChild("detection"),
            output_cleaners=self.get_output_cleaners(config),
            processing_filters=self.get_processing_filters(config),
            message_factories=self.get_message_factories(config),
        )

    def get_recording_callbacks(self, config: C) -> List[Callable]:
        """Get the recording callbacks.

        This method adds the detection task as a callback to be executed
        after each recording is completed.

        Parameters
        ----------
        config : C
            The program's configuration.

        Returns
        -------
        List[Callable]
            A list of recording callbacks, including the detection task.
        """
        return [self.create_detection_task(config)]

    def get_required_models(self, config: C) -> list[str]:
        name = getattr(self.model, "name", None)
        if name is not None:
            return [name]

        name = getattr(self.model, "__name__", None)
        if name is not None:
            return [name]

        raise ValueError("Model must have a name attribute.")



"""System functions from managing program config files."""

import json
from pathlib import Path
from typing import (
    Any,
    List,
    Optional,
    Type,
    TypeVar,
    get_args,
    get_origin,
)

from omegaconf import OmegaConf
from pydantic import BaseModel, ValidationError

from acoupi.system import exceptions

__all__ = [
    "dump_config",
    "get_config_field",
    "load_config",
    "set_config_field",
    "write_config",
]


S = TypeVar("S", bound=BaseModel)


def write_config(
    config: Optional[BaseModel],
    path: Path,
) -> None:
    """Write config to file."""
    if config is None:
        return

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    with open(path, "w") as file:
        file.write(config.model_dump_json(indent=2))


def load_config(path: Path, schema: Type[S]) -> S:
    """Load config from file.

    Parameters
    ----------
    path
        Path to the config file.
    schema
        Pydantic model to validate the config.

    Returns
    -------
    S
        The loaded config.

    Raises
    ------
    ConfigurationError
        If the configuration is invalid.
    FileNotFoundError
        If the config file does not exist.
    """
    with open(path) as file:
        try:
            return schema.model_validate_json(file.read())
        except (ValueError, ValidationError) as error:
            raise exceptions.ConfigurationError(
                message=f"Invalid configuration in {path}: {error}",
                help="Check the configuration file for errors.",
            ) from error


def get_config_field(config: BaseModel, field: str) -> Any:
    """Retrieve a field or nested field from a configuration object.

    This function allows you to access configuration values by specifying the
    field name, including nested fields using dot notation
    (e.g., 'section.subsection.value'). If no field is provided, the entire
    configuration object is returned.

    It is possible to get nested fields by using a dot notation.

    Parameters
    ----------
    config
        The configuration object, an instance of a Pydantic model.
    field
        The name of the field to retrieve. Use dot notation for nested fields.

    Returns
    -------
    Any
        The value of the specified field. The type depends on the field's
        definition in the Pydantic model.

    Raises
    ------
    AttributeError
        If the specified `field` does not exist within the configuration schema.
    IndexError
        If the `field` refers to an index in a list or tuple that is out of
        bounds.
    NotImplementedError
        If the `field` points to a data type or structure that this function
        does not currently support for retrieval.

    Examples
    --------
    Accessing a root-level field:

    >>> class Config(BaseModel):
    ...     a: int
    ...     b: str
    >>> config = Config(a=1, b="2")
    >>> get_config_field(config, "a")
    >>> 1

    Accessing a nested field:

    >>> class NestedConfig(BaseModel):
    ...     d: bool
    >>> class Config(BaseModel):
    ...     a: int
    ...     b: str
    ...     c: NestedConfig
    >>> config = Config(a=1, b="2", c=NestedConfig(d=True))
    >>> get_config_field(config, "c.d")
    >>> True

    Accessing an element from a list:

    >>> class Config(BaseModel):
    ...     a: List[int]
    >>> config = Config(a=[1, 2, 3])
    >>> get_config_field(config, "a.0")
    >>> 1
    """
    if field is None or field == "":
        return config

    fields = type(config).model_fields
    prefix, *rest = field.split(".", 1)

    if prefix not in fields:
        valid_fields = "\n\t - ".join(fields.keys())
        raise AttributeError(
            f"Field '{prefix}' is not part of the configuration schema.\n"
            f"Valid fields are:\n\t - {valid_fields}",
        )

    subfield = fields[prefix]
    type_ = subfield.rebuild_annotation()
    origin = get_origin(type_)

    if not rest:
        return getattr(config, prefix)

    if origin is not None and issubclass(origin, (tuple, list)):
        return _get_config_field_from_sequence(
            getattr(config, prefix),
            rest[0],
            type_,
        )

    if issubclass(type_, BaseModel):
        return get_config_field(getattr(config, prefix), rest[0])

    raise NotImplementedError(f"Cannot get field from {type_}.")


def set_config_field(
    config: S,
    field: str,
    value: Any,
    is_json: bool = False,
    strict: bool = False,
    from_attributes: bool = False,
) -> S:
    """Set a specific field in a config object.

    This function enables you to update configuration values by specifying the
    field name, including nested fields using dot notation
    (e.g., 'section.subsection.value'). The provided `value` is validated
    against the field's type definition in the Pydantic model to ensure data
    integrity.

    Parameters
    ----------
    config
        The configuration object to be modified. Must be an instance of a
        Pydantic model.
    field
        The name of the field to set. Use dot notation for nested fields. If an
        empty string, the entire configuration is replaced. This is useful when
        setting the configuration from a JSON string.
    value
        The new value to assign to the specified field. The type should be
        compatible with the field's definition in the Pydantic model.
    is_json
        If True, the `value` is treated as a JSON string and parsed before
        being assigned. Default is False.
    strict
        If True, type coercion is disabled, and the `value` must match the
        field's type exactly. If False (default), type coercion is attempted if
        possible.
    from_attributes
        Relevant only when setting a field that is itself a Pydantic model. If
        True, the `value` is assumed to be an object whose attributes will be
        used to populate the model. See Pydantic's `model_validate`
        documentation for details. Default is False.

    Returns
    -------
    S
        A new configuration object with the specified field modified. The
        original `config` object remains unchanged.

    Raises
    ------
    ValidationError
        If the provided `value` fails validation against the field's type
        definition in the Pydantic model.
    AttributeError
        If the specified `field` does not exist within the configuration
        schema.
    IndexError
        If the `field` refers to an index in a list or tuple that is out of
        bounds.
    NotImplementedError
        If the `field` points to a data type or structure that this function
        does not currently support for setting.

    Examples
    --------
    Setting a root-level field:

    >>> class Config(BaseModel):
    ...     a: int
    ...     b: str
    >>> config = Config(a=1, b="2")
    >>> new_config = set_config_field(config, "a", 3)
    >>> new_config.a
    >>> 3

    Setting a nested field:

    >>> class NestedConfig(BaseModel):
    ...     d: bool
    >>> class Config(BaseModel):
    ...     a: int
    ...     b: str
    ...     c: NestedConfig
    >>> config = Config(a=1, b="2", c=NestedConfig(d=True))
    >>> new_config = set_config_field(config, "c.d", False)
    >>> new_config.c.d

    Setting an element in a list:

    >>> class Config(BaseModel):
    ...     a: List[int]
    >>> config = Config(a=[1, 2, 3])
    >>> new_config = set_config_field(config, "a.0", 4)
    >>> new_config.a[0]
    >>> 4

    Setting a field from a JSON string:

    >>> class Config(BaseModel):
    ...     a: int
    ...     b: str
    >>> config = Config(a=1, b="2")
    >>> new_config = set_config_field(
    ...     config,
    ...     "",
    ...     '{"a": 3, "b": "4"}',
    ...     is_json=True,
    ... )
    >>> new_config.a
    >>> 3

    """
    if not field:
        if isinstance(value, type(config)):
            return value

        if is_json:
            return config.model_validate_json(value, strict=strict)

        return config.model_validate(
            value,
            strict=strict,
            from_attributes=from_attributes,
        )

    base = OmegaConf.create(config.model_dump(), flags={"allow_objects": True})

    if is_json:
        value = json.loads(value)

    OmegaConf.update(base, field, value)

    try:
        return type(config).model_validate(base, extra="forbid")
    except ValidationError as error:
        raise exceptions.ParameterError(
            value=field,
            message=f"Invalid field {field} in {type(config)}.",
            help="Check the field name and the type of the value.",
        ) from error


class PydanticJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder for Pydantic models."""

    def default(self, o: Any) -> Any:
        if isinstance(o, BaseModel):
            return json.loads(o.model_dump_json(round_trip=True))

        return super().default(o)


def dump_config(config: Any, indent: int = 2) -> str:
    """Dump a configuration object to a JSON string."""
    return json.dumps(config, cls=PydanticJSONEncoder, indent=indent)


def _get_config_field_from_sequence(
    config: List[Any],
    field: str,
    type_: Type,
) -> Any:
    prefix, *rest = field.split(".", 1)

    if not prefix.isdigit():
        raise ValueError("Field is not an index.")

    index = int(prefix)
    if index >= len(config) or index < 0:
        raise IndexError("Index out of range.")

    if not rest:
        return config[index]

    args = get_args(type_)

    if not args:
        raise NotImplementedError(
            "Cannot get field from list of non-models.",
        )

    origin = get_origin(type_)
    assert origin is not None

    if issubclass(origin, list):
        type_ = args[0]
    elif issubclass(origin, tuple):
        type_ = args[index]
    else:
        raise NotImplementedError(f"Cannot get field from {origin}.")

    if issubclass(type_, BaseModel):
        return get_config_field(config[index], rest[0])

    raise NotImplementedError(f"Cannot get field from list of {args[0]}")



"""Test Summariser."""

import datetime
from pathlib import Path
from typing import List

import pytest

from acoupi import data
from acoupi.components import summariser


@pytest.fixture
def create_test_model_output():
    """Return a model output."""
    deployment = data.Deployment(
        name="test_deployment",
    )

    recording = data.Recording(
        path=Path("test.wav"),
        duration=1,
        samplerate=256000,
        audio_channels=1,
        deployment=deployment,
        created_on=datetime.datetime.now(),
    )

    def factory(
        tags: List[data.PredictedTag],
    ) -> data.ModelOutput:
        """Return a model output."""
        return data.ModelOutput(
            recording=recording,
            name_model="test_model",
            tags=tags,
        )

    return factory


@pytest.fixture
def create_test_predictedtags():
    """Fixture for creating random predicted tags.

    Will create multiple random tags.
    """

    def factory(
        tag_key: str = "species",
    ) -> List[data.PredictedTag]:
        """Return a random detection."""
        return [
            data.PredictedTag(
                tag=data.Tag(
                    key=tag_key,
                    value="specie_a",
                ),
                confidence_score=0.45,
            ),
            data.PredictedTag(
                tag=data.Tag(
                    key=tag_key,
                    value="specie_b",
                ),
                confidence_score=0.65,
            ),
            data.PredictedTag(
                tag=data.Tag(
                    key=tag_key,
                    value="specie_c",
                ),
                confidence_score=0.85,
            ),
        ]

    return factory

    def test_get_species_name(
        create_test_model_output,
        create_test_predictedtags,
    ):
        """Test get species name."""
        model_output = create_test_model_output(
            tags=create_test_predictedtags(),
        )
        summariser_tester = summariser.Summariser(
            threshold_lowband=0.5,
            threshold_midband=0.7,
            threshold_highband=0.9,
        )
        species_name = summariser_tester.get_species_name(
            model_outputs=[model_output],
        )
        assert species_name == ["specie_a", "specie_b", "specie_c"]

    def test_get_species_count_in_bands(
        create_test_model_output,
        create_test_predictedtags,
    ):
        """Test get species count in bands."""
        model_output = create_test_model_output(
            tags=create_test_predictedtags(),
        )
        summariser_tester = summariser.Summariser(
            threshold_lowband=0.5,
            threshold_midband=0.7,
            threshold_highband=0.9,
        )
        species_count = summariser_tester.get_species_count_in_bands(
            species_name="specie_c",
            model_outputs=[model_output],
        )
        assert species_count == {
            "low_band": 0,
            "mid_band": 0,
            "high_band": 1,
        }

        def test_build_summary(
            create_test_model_output,
            create_test_predictedtags,
        ):
            """Test build summary."""
            model_output = create_test_model_output(
                tags=create_test_predictedtags(),
            )
            summariser_tester = summariser.Summariser(
                threshold_lowband=0.5,
                threshold_midband=0.7,
                threshold_highband=0.9,
            )
            summary = summariser_tester.build_summary(
                model_outputs=[model_output],
            )
            assert summary == data.Message(
                species_count={
                    "specie_a": {
                        "low_band": 0,
                        "mid_band": 0,
                        "high_band": 0,
                    },
                    "specie_b": {
                        "low_band": 1,
                        "mid_band": 0,
                        "high_band": 0,
                    },
                    "specie_c": {
                        "low_band": 0,
                        "mid_band": 0,
                        "high_band": 1,
                    },
                },
            )



"""Test the SQLite store."""

import datetime
import sqlite3
import uuid
from pathlib import Path
from typing import Generator, cast

import pytest

from acoupi import components, data
from acoupi.components.stores.sqlite import store as sqlite_store_module
from acoupi.system.exceptions import MetadataStoreError


@pytest.fixture(scope="function")
def sqlite_store(
    tmp_path: Path,
) -> Generator[components.SqliteStore, None, None]:
    """Create a store."""
    db_path = tmp_path / "test.db"
    yield components.SqliteStore(db_path)
    db_path.unlink()


def test_sqlite_store_creates_a_database_file(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the store creates a database file."""
    db_path = sqlite_store.db_path
    assert db_path.exists()


def test_database_has_correct_tables(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the database has the correct tables."""
    expected_tables = {
        "recording",
        "deployment",
        "detection",
    }
    db_path = sqlite_store.db_path

    # Check that the tables are there
    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        actual_tables = set(row[0] for row in cursor.fetchall())
        assert expected_tables.issubset(actual_tables)


def test_recording_table_has_correct_columns(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the recording table has the correct columns."""
    expected_columns = {
        "id",
        "path",
        "duration_s",
        "samplerate_hz",
        "audio_channels",
        "datetime",
        "deployment_id",
    }
    db_path = sqlite_store.db_path

    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("PRAGMA table_info(recording);")
        actual_columns = set(row[1] for row in cursor.fetchall())
        assert expected_columns == actual_columns


def test_deployment_table_has_correct_columns(sqlite_store) -> None:
    """Test that the deployment table has the correct columns."""
    expected_columns = {
        "id",
        "started_on",
        "ended_on",
        "latitude",
        "longitude",
        "name",
    }
    db_path = sqlite_store.db_path

    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("PRAGMA table_info(deployment);")
        actual_columns = set(row[1] for row in cursor.fetchall())
        assert expected_columns == actual_columns


def test_detection_table_has_correct_columns(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the detection table has the correct columns."""
    expected_columns = {
        "id",
        "start_time_s",
        "end_time_s",
        "low_freq_hz",
        "high_freq_hz",
        "detection_score",
        "model_output_id",
    }
    db_path = sqlite_store.db_path

    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("PRAGMA table_info(detection);")
        actual_columns = set(row[1] for row in cursor.fetchall())
        assert expected_columns == actual_columns


def test_can_instantiate_multiple_stores_with_the_same_sqlite_file(
    tmp_path: Path,
) -> None:
    """Test that we can instantiate multiple stores to the same sqlite file."""
    # Arrange
    db_path = tmp_path / "test.db"

    # Act
    store1 = components.SqliteStore(db_path)
    store2 = components.SqliteStore(db_path)

    # Assert
    assert store1 is not None
    assert store2 is not None


def test_can_store_a_deployment(sqlite_store: components.SqliteStore) -> None:
    """Test that we can store a deployment."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )

    # Act
    sqlite_store.store_deployment(deployment)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM deployment;")
        row = cursor.fetchone()
        assert row["started_on"] == now.isoformat(sep=" ")
        assert row["latitude"] == 1.0
        assert row["longitude"] == 2.0


def test_get_current_deployment_returns_new_if_no_deployment_exists(
    sqlite_store: components.SqliteStore,
    patched_now,
) -> None:
    """Test that get_current_deployment fails if no deployment exists."""
    # Arrange
    # Patch the current time
    now = patched_now()

    # Make sure there are no deployments
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("DELETE FROM deployment;")

    # Act
    # Try to get the current deployment, this should create a new one
    deployment = sqlite_store.get_current_deployment()

    # Assert
    assert deployment.latitude is None
    assert deployment.longitude is None
    assert deployment.started_on == now


def test_get_current_deployment_gets_latest_deployment(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that get_current_deployment gets the latest deployment."""
    # Arrange
    start_1 = datetime.datetime.now()
    start_2 = datetime.datetime.now() - datetime.timedelta(days=10)
    deployment1 = data.Deployment(
        started_on=start_1,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    deployment2 = data.Deployment(
        started_on=start_2,
        latitude=3.0,
        longitude=4.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment1)
    sqlite_store.store_deployment(deployment2)

    # Act
    deployment = sqlite_store.get_current_deployment()

    # Assert
    assert deployment == deployment1


def test_deployment_id_in_db_is_same_as_in_python(
    sqlite_store: components.SqliteStore,
) -> None:
    """Test that the deployment_id in the db is the same as in python."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )

    # Act
    sqlite_store.store_deployment(deployment)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM deployment;")
        row = cursor.fetchone()
        assert row["id"] == deployment.id.bytes


def test_recordings_can_be_registered(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
) -> None:
    """Test that recordings can be registered."""
    # Arrange
    now = datetime.datetime.now()
    recording = data.Recording(
        deployment=deployment,
        path=Path("test/path"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
        audio_channels=2,
    )

    # Act
    sqlite_store.store_recording(recording)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM recording;")
        row = cursor.fetchone()
        assert row["path"] == "test/path"
        assert row["duration_s"] == 10.0
        assert row["audio_channels"] == 2
        assert row["samplerate_hz"] == 44100
        assert row["datetime"] == now.isoformat(sep=" ")
        assert row["id"] == recording.id.bytes


def test_recording_can_be_registered_with_custom_deployment(
    sqlite_store: components.SqliteStore,
):
    """Test that recordings can be registered with a custom deployment."""
    # Arrange
    now = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=now,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment)
    recording = data.Recording(
        path=Path("test/path"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
        deployment=deployment,
    )

    # Act
    sqlite_store.store_recording(recording)

    # Assert
    with sqlite3.connect(str(sqlite_store.db_path)) as conn:
        conn.row_factory = sqlite3.Row
        cursor = conn.cursor()
        cursor.execute("SELECT * FROM recording;")
        row = cursor.fetchone()
        assert row["path"] == "test/path"
        assert row["duration_s"] == 10.0
        assert row["samplerate_hz"] == 44100
        assert row["datetime"] == now.isoformat(sep=" ")
        assert row["id"] == recording.id.bytes
        assert row["deployment_id"] == deployment.id.bytes


def test_get_all_recordings(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
):
    """Test that all recordings can be retrieved."""
    # Arrange

    # Create two recordings
    now = datetime.datetime.now()
    recording1 = data.Recording(
        deployment=deployment,
        path=Path("test/path1"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
    )
    recording2 = data.Recording(
        deployment=deployment,
        path=Path("test/path2"),
        duration=10.0,
        samplerate=44100,
        created_on=now + datetime.timedelta(seconds=10),
    )

    # Add them to the database
    sqlite_store.store_recording(recording1)
    sqlite_store.store_recording(recording2)

    # Act
    recordings = sqlite_store.get_recordings([recording1.id, recording2.id])

    # Assert
    assert len(recordings) == 2
    assert recordings[1][0] == recording1
    assert recordings[0][0] == recording2


def test_get_some_recordings(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
):
    """Test that recordings can be retrieved with an include filter."""
    # Arrange

    # Create two recordings
    now = datetime.datetime.now()
    recording1 = data.Recording(
        deployment=deployment,
        path=Path("test/path1"),
        duration=10.0,
        samplerate=44100,
        created_on=now,
    )
    recording2 = data.Recording(
        deployment=deployment,
        path=Path("test/path2"),
        duration=10.0,
        samplerate=44100,
        created_on=now + datetime.timedelta(seconds=10),
    )

    # Add them to the database
    sqlite_store.store_recording(recording1)
    sqlite_store.store_recording(recording2)

    # Act
    recordings = sqlite_store.get_recordings([recording1.id])

    # Assert
    assert len(recordings) == 1
    assert recordings[0][0] == recording1


def test_can_store_model_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_model_output(model_output)
    db_path = sqlite_store.db_path

    retrieved = sqlite_store.get_recordings([model_output.recording.id])
    assert len(retrieved) == 1

    _, model_outputs = retrieved[0]
    assert len(model_outputs) == 1

    retrieved = model_outputs[0]

    assert len(retrieved.detections) == len(model_output.detections)
    assert retrieved == model_output

    # Check that the detections were stored
    with sqlite3.connect(str(db_path)) as conn:
        cursor = conn.cursor()
        cursor.execute("SELECT id FROM model_output;")
        assert len(cursor.fetchall()) == 1

        cursor.execute("SELECT * FROM detection;")
        row = cursor.fetchone()
        assert row is not None
        assert row[0] == model_output.detections[0].id.bytes
        assert row[1] == 1
        assert row[2] == 2
        assert row[3] == 1000
        assert row[4] == 2000
        assert row[5] == model_output.detections[0].detection_score


def test_can_store_multiple_model_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    second_recording = model_output.recording.model_copy(
        update=dict(
            id=uuid.uuid4(),
            path=Path("test/path-second"),
            created_on=model_output.recording.created_on
            + datetime.timedelta(seconds=5),
        )
    )
    second_output = model_output.model_copy(
        update=dict(
            id=uuid.uuid4(),
            recording=second_recording,
            created_on=model_output.created_on + datetime.timedelta(seconds=5),
            detections=[
                detection.model_copy(update=dict(id=uuid.uuid4()))
                for detection in model_output.detections
            ],
        )
    )

    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_recording(second_recording)
    sqlite_store.store_model_outputs([model_output, second_output])

    retrieved = sqlite_store.get_recordings(
        [model_output.recording.id, second_output.recording.id]
    )
    assert len(retrieved) == 2


def test_get_recordings_by_path_returns_full_model_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_model_output(model_output)

    retrieved = sqlite_store.get_recordings_by_path(
        [model_output.recording.path]
    )

    assert len(retrieved) == 1
    recording, outputs = retrieved[0]
    assert recording == model_output.recording
    assert outputs == [model_output]


def test_get_recordings_info_by_path_returns_lightweight_model_output_info(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_model_output(model_output)

    retrieved = sqlite_store.get_recordings_info_by_path(
        [model_output.recording.path]
    )

    assert len(retrieved) == 1
    recording, outputs = retrieved[0]
    assert recording == model_output.recording
    assert len(outputs) == 1
    assert outputs[0].id == model_output.id
    assert outputs[0].name_model == model_output.name_model
    assert outputs[0].created_on == model_output.created_on


def test_get_recording_model_outputs_returns_full_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_model_output(model_output)

    retrieved = sqlite_store.get_recording_model_outputs(
        model_output.recording
    )

    assert retrieved == [model_output]


def test_get_recordings_model_outputs_returns_outputs_grouped_by_recording(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    second_recording = model_output.recording.model_copy(
        update=dict(
            id=uuid.uuid4(),
            path=Path("test/path-second"),
            created_on=model_output.recording.created_on
            + datetime.timedelta(seconds=5),
        )
    )
    second_output = model_output.model_copy(
        update=dict(
            id=uuid.uuid4(),
            recording=second_recording,
            created_on=model_output.created_on + datetime.timedelta(seconds=5),
            detections=[
                detection.model_copy(update=dict(id=uuid.uuid4()))
                for detection in model_output.detections
            ],
        )
    )

    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_recording(second_recording)
    sqlite_store.store_model_outputs([model_output, second_output])

    retrieved = sqlite_store.get_recordings_model_outputs(
        [model_output.recording, second_recording]
    )

    assert retrieved == {
        model_output.recording.id: [model_output],
        second_recording.id: [second_output],
    }


def test_get_recordings_model_outputs_returns_empty_list_for_recording_without_outputs(
    sqlite_store: components.SqliteStore,
    recording: data.Recording,
):
    sqlite_store.store_recording(recording)

    retrieved = sqlite_store.get_recordings_model_outputs([recording])

    assert retrieved == {recording.id: []}


def test_get_recordings_model_outputs_includes_requested_recordings_without_outputs(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    recording_without_outputs = model_output.recording.model_copy(
        update=dict(
            id=uuid.uuid4(),
            path=Path("test/path-without-outputs"),
            created_on=model_output.recording.created_on
            + datetime.timedelta(seconds=5),
        )
    )

    sqlite_store.store_recording(model_output.recording)
    sqlite_store.store_recording(recording_without_outputs)
    sqlite_store.store_model_output(model_output)

    retrieved = sqlite_store.get_recordings_model_outputs(
        [model_output.recording, recording_without_outputs]
    )

    assert retrieved == {
        model_output.recording.id: [model_output],
        recording_without_outputs.id: [],
    }


def test_get_recordings_model_outputs_chunks_large_batches(
    sqlite_store: components.SqliteStore,
    deployment: data.Deployment,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        sqlite_store_module,
        "SQLITE_MAX_BOUND_VARIABLES",
        2,
    )

    recordings = []
    model_outputs = []
    base_time = datetime.datetime.now()
    for index in range(5):
        recording = data.Recording(
            deployment=deployment,
            path=Path(f"test/chunked-{index}.wav"),
            duration=10.0,
            samplerate=44100,
            created_on=base_time + datetime.timedelta(seconds=index),
        )
        model_output = data.ModelOutput(
            name_model=f"test_model_{index}",
            recording=recording,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(key="test", value=f"value-{index}"),
                    confidence_score=0.8,
                )
            ],
            detections=[
                data.Detection(
                    id=uuid.uuid4(),
                    location=data.BoundingBox(coordinates=(1, 1000, 2, 2000)),
                    detection_score=0.6,
                    tags=[
                        data.PredictedTag(
                            tag=data.Tag(key="species", value=f"tag-{index}"),
                            confidence_score=0.3,
                        )
                    ],
                )
            ],
            created_on=recording.created_on,
        )
        recordings.append(recording)
        model_outputs.append(model_output)

    for recording in recordings:
        sqlite_store.store_recording(recording)
    sqlite_store.store_model_outputs(model_outputs)

    retrieved = sqlite_store.get_recordings_model_outputs(recordings)

    assert retrieved == {
        recording.id: [model_output]
        for recording, model_output in zip(
            recordings,
            model_outputs,
            strict=True,
        )
    }


def test_store_model_outputs_fails_if_recording_is_missing(
    sqlite_store: components.SqliteStore,
    model_output: data.ModelOutput,
):
    with pytest.raises(MetadataStoreError) as excinfo:
        sqlite_store.store_model_outputs([model_output])

    message = str(excinfo.value)
    assert "not present in the metadata store" in message
    assert str(model_output.recording.id) in message
    assert str(model_output.recording.path) in message
    assert "store_recording()" in message


def test_in_memory_store_raises_clear_error_for_batched_model_output_loads(
    deployment: data.Deployment,
):
    sqlite_store = components.SqliteStore(cast(Path, ":memory:"))
    recording = data.Recording(
        deployment=deployment,
        path=Path("test/in-memory.wav"),
        duration=10.0,
        samplerate=44100,
        created_on=datetime.datetime.now(),
    )

    with pytest.raises(MetadataStoreError) as excinfo:
        sqlite_store.get_recordings_model_outputs([recording])

    assert "Direct sqlite batch operations are not supported" in str(
        excinfo.value
    )
    assert "file-backed sqlite store" in str(excinfo.value)


def test_can_update_deployment_info(
    sqlite_store: components.SqliteStore,
):
    start = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=start,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    sqlite_store.store_deployment(deployment)

    current_deployment = sqlite_store.get_current_deployment()
    assert deployment == current_deployment

    assert current_deployment.ended_on is None

    updated_deployment = deployment.model_copy(
        update=dict(ended_on=start + datetime.timedelta(days=2))
    )
    assert updated_deployment.ended_on is not None
    sqlite_store.update_deployment(updated_deployment)

    current_deployment = sqlite_store.get_current_deployment()
    assert updated_deployment.model_dump() == current_deployment.model_dump()
    assert current_deployment.ended_on is not None


def test_update_deployment_fails_if_corresponding_deployment_does_not_exist(
    sqlite_store: components.SqliteStore,
):
    start = datetime.datetime.now()
    deployment = data.Deployment(
        started_on=start,
        latitude=1.0,
        longitude=2.0,
        name="test_device",
    )
    with pytest.raises(ValueError):
        sqlite_store.update_deployment(deployment)



import datetime
import logging
from typing import Callable, List

from pydantic import BaseModel, Field

from acoupi.components import types
from acoupi.data import Message
from acoupi.devices import get_device_id

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Heartbeat(BaseModel):
    sent_on: datetime.datetime = Field(default_factory=datetime.datetime.now)
    device_id: str
    status: str = "OK"


def generate_heartbeat_task(
    messengers: List[types.Messenger],
    logger: logging.Logger = logger,
) -> Callable[[], None]:
    """Generate a heartbeat task.

    The heartbeat task creates a heartbeat message and emits them using the
    provided messengers.

    Parameters
    ----------
    messengers : Optional[List[types.Messenger]]
        A list of messenger instances to pass the heartbeat message to.
    logger : logging.Logger
        Logger instance for logging heartbeat status.

    Returns
    -------
    Callable[[], None]
        A function that sends a heartbeat message when called.
    """
    if not messengers:
        raise ValueError(
            "At least one messenger is required to send a heartbeat message."
        )

    def heartbeat_task() -> None:
        device_id = get_device_id()

        now = datetime.datetime.now()

        heartbeat = Heartbeat(device_id=device_id, sent_on=now)
        message = Message(
            content=heartbeat.model_dump_json(),
            created_on=now,
        )

        for messenger in messengers:
            response = messenger.send_message(message)
            logger.info(f"Heartbeat Sent - Response Status: {response.status}")

    return heartbeat_task



import logging
from pathlib import Path
from typing import Optional

import click

from acoupi import system
from acoupi.cli.base import acoupi

__all__ = [
    "task",
]


@acoupi.group()
@click.pass_context
def task(ctx):
    """Manage acoupi tasks.

    Provides commands to list, run, and profile tasks within
    your configured acoupi program.
    """
    settings = ctx.obj["settings"]
    if not system.is_configured(settings):
        click.echo("Acoupi is not setup. Run `acoupi setup` first.")
        return


@task.command()
@click.pass_context
def list(ctx):
    """List all available tasks in the current acoupi program."""
    program = system.load_program(ctx.obj["settings"])
    task_list = system.get_task_list(program)

    click.echo(
        "Program "
        + click.style(type(program).__name__, fg="green", bold=True)
        + " tasks:\n"
    )

    for task in task_list:
        click.secho(f"  • {task}", bold=True)


@task.command()
@click.argument("task_name", type=str)
@click.pass_context
def run(ctx, task_name: str):
    """Run a specified task.

    Parameters
    ----------
    task_name : str
        The name of the task to run.
    """
    program = system.load_program(ctx.obj["settings"])
    task_list = system.get_task_list(program)

    if task_name not in task_list:
        raise click.UsageError(
            click.style("Task ", fg="red")
            + click.style(task_name, fg="red", bold=True)
            + click.style(" not found.", fg="red")
        )

    click.echo(f"Running task {click.style(task_name, fg='green')}")
    system.run_task(program, task_name)


@task.command()
@click.argument("task_name", type=str)
@click.option(
    "--output",
    type=Path,
    help="Path to save profiling output.",
)
@click.option(
    "--quiet",
    type=bool,
    default=False,
    help="Suppress printing profiling output to the console.",
)
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
    default="WARNING",
    help="Set the logging level.",
)
@click.pass_context
def profile(
    ctx,
    task_name: str,
    output: Optional[Path],
    quiet: bool,
    log_level: str,
):
    """Profile a specified task.

    Parameters
    ----------
    task_name : str
        The name of the task to profile.
    """
    program = system.load_program(ctx.obj["settings"])
    task_list = system.get_task_list(program)

    if not quiet:
        logging.basicConfig(level=log_level)

    if task_name not in task_list:
        raise click.UsageError(
            click.style("Task ", fg="red")
            + click.style(task_name, fg="red", bold=True)
            + click.style(" not found.", fg="red")
        )

    if not quiet:
        click.echo(f"Profiling task {click.style(task_name, fg='green')}")

    stats = system.profile_task(program, task_name)

    if output:
        stats.dump_stats(output)



"""Recording Schedulers for acoupi.

Recording schedulers are used to determine how often recordings should be made.
This is useful for example if you want to record at a constant interval, or if
you want to record at a variable interval, such as every 10 minutes during the
day and every 30 minutes at night.

Recording schedulers are implemented as classes that inherit from
RecordingScheduler. The class should implement the time_until_next_recording
method, which returns the time in seconds until the next recording should be
made.
"""

import datetime
from typing import Optional

from acoupi.components.types import RecordingScheduler

__all__ = [
    "IntervalScheduler",
]


class IntervalScheduler(RecordingScheduler):
    """Will wait for a constant amount of time between each recording."""

    interval: float
    """The interval between each recording. In seconds."""

    def __init__(self, timeinterval: float):
        """Initialise the recording scheduler.

        Parameters
        ----------
        timeinterval: float
            The interval between each recording. In seconds.
        """
        self.timeinterval = timeinterval

    def time_until_next_recording(
        self, time: Optional[datetime.datetime] = None
    ) -> float:
        """Provide the number of seconds until the next recording.

        Parameters
        ----------
        time : Optional[datetime.datetime], optional
            The time to use for determining the next recording, by default None.

        Returns
        -------
        float
            The number of seconds until the next recording.
            Will return 0 if a recording should be made immediately.
        """
        if not time:
            time = datetime.datetime.now()
        next_recording_time = (
            time + datetime.timedelta(seconds=self.timeinterval)
        ).replace(microsecond=0)
        return int((next_recording_time - time).total_seconds())



"""Test suite for acoupi program system module."""

import sys
from pathlib import Path

import pytest

from acoupi.programs.default import Program as SampleProgram
from acoupi.programs.test import TestProgram
from acoupi.system import Settings, exceptions, programs


def test_write_program_file_creates_a_file(settings: Settings):
    """Test write_program_file creates a file."""
    program_name = "test_program"
    programs.write_program_file(program_name, settings)
    program_file = settings.program_file
    assert program_file.exists()


def test_write_program_file_has_correct_program_name(settings: Settings):
    """Test write_program_file has correct program name."""
    program_name = "test_program"
    programs.write_program_file(
        program_name,
        settings,
    )

    program_name_file = settings.program_name_file
    program_file = settings.program_file
    line = f'program_name_file=Path("{str(program_name_file)}")'
    with open(program_file) as file:
        assert line in file.read()


def test_write_program_file_points_to_correct_config_file(settings: Settings):
    """Test write_program_file points to correct config file."""
    program_name = "test_program"
    programs.write_program_file(program_name, settings)

    config_file = settings.program_config_file
    program_file = settings.program_file

    line = f'config_file=Path("{config_file}")'
    with open(program_file) as file:
        assert line in file.read()


def test_can_load_the_test_program():
    """Test can load the test program."""
    program_class = programs.load_program_class("acoupi.programs.test")
    assert program_class == TestProgram


def test_can_load_the_acoupi_program():
    """Test can load the test program."""
    program_class = programs.load_program_class("acoupi.programs.default")
    assert program_class == SampleProgram


sample_program = """
from acoupi.programs import AcoupiProgram
from acoupi.programs.templates import BasicProgram, BasicProgramConfiguration


class Config(BasicProgramConfiguration):
    name: str = "test"


class Program(BasicProgram[Config]):
    config_schema = Config

    name = "test_program"
"""


def test_loads_correct_program_when_multiple_programs_exist(tmp_path: Path):
    program_file = tmp_path / "test_program.py"
    program_file.write_text(sample_program)
    sys.path.append(str(tmp_path))
    program_class = programs.load_program_class("test_program")
    assert program_class.__name__ == "Program"
    assert hasattr(program_class, "name")
    assert getattr(program_class, "name") == "test_program"


def test_load_program_fails_if_module_does_not_exist():
    """Test load_program fails if file does not exist."""
    with pytest.raises(exceptions.ProgramNotFoundError):
        programs.load_program_class("does.not.exist")


def test_load_program_fails_if_no_program_is_found_in_module():
    """Test load_program fails if no program is found in module."""
    with pytest.raises(exceptions.ProgramNotFoundError):
        programs.load_program_class("math")



"""System functions for managing Acoupi programs."""

import inspect
import warnings
from enum import Enum
from importlib import import_module
from typing import Type

from celery import Celery
from pydantic import BaseModel

from acoupi.programs.core import base as programs
from acoupi.programs.core.workers import DEFAULT_WORKER_CONFIG, WorkerConfig
from acoupi.system import exceptions
from acoupi.system.config import load_config
from acoupi.system.constants import CeleryConfig, Settings
from acoupi.system.templates import render_template

__all__ = [
    "load_program",
    "load_program_class",
    "load_config_schema",
    "write_program_file",
]


def load_program_class(program_name: str) -> Type[programs.AcoupiProgram]:
    """Load the acoupi program class from a specified module.

    This function searches the given module for a valid AcoupiProgram class. If
    multiple such classes exist within the module, it prioritizes classes that
    are not further subclassed.

    Parameters
    ----------
    program_name
        The name of the module containing the acoupi program class.

    Returns
    -------
    Type[programs.AcoupiProgram]
        The acoupi program class.

    Raises
    ------
    exceptions.ProgramNotFoundError
        If the specified program module is not found.
    exceptions.InvalidProgramError
        If the loaded class is not a valid acoupi program class.

    Notes
    -----
    When multiple AcoupiProgram classes are found within the module, this
    function will attempt to select one that is not further subclassed. If
    multiple such "final" classes exist, a warning will be issued, and the
    first one encountered will be selected. It is recommended that when
    exposing AcoupiProgram classes in a module, only one "final"
    (non-subclassable) class be made available.
    """
    try:
        program_module = import_module(program_name)
    except ModuleNotFoundError as e:
        raise exceptions.ProgramNotFoundError(program=program_name) from e

    def is_acoupi_program_class(cls):
        return (
            inspect.isclass(cls)
            and not inspect.isabstract(cls)
            and issubclass(cls, programs.AcoupiProgram)
        )

    all_programs = inspect.getmembers(program_module, is_acoupi_program_class)

    if not all_programs:
        raise exceptions.ProgramNotFoundError(program=program_name)

    valid_classes = dict(all_programs)

    for _, class_ in all_programs:
        assert issubclass(class_, programs.AcoupiProgram)

        for base_class in class_.__bases__:
            base_name = base_class.__name__
            valid_classes.pop(base_name, None)

    if len(valid_classes) > 1:
        warnings.warn(
            f"Found multiple programs in {program_name}. "
            "Imported the first one but this may not be the desired behavior.",
            stacklevel=1,
        )

    if valid_classes:
        return valid_classes.popitem()[1]

    raise exceptions.InvalidProgramError(program=program_name)


def load_program(settings: Settings) -> programs.AcoupiProgram:
    """Load a fully configured acoupi program.

    This function loads the acoupi program class from the specified module,
    instantiates it, and configures it with the specified config files.

    Parameters
    ----------
    settings
        The settings object containing the paths to the program and celery
        config files.

    Returns
    -------
    programs.AcoupiProgram
        A fully configured instance of the acoupi program class.

    Raises
    ------
    exceptions.ProgramNotFoundError
        If the program module is not found.
    exceptions.InvalidProgramError
        If the loaded class is not a valid acoupi program class.
    """
    program_name = settings.program_name_file.read_text().strip()
    program_class = load_program_class(program_name)
    config_schema = program_class.get_config_schema()

    if config_schema is not None:
        config = load_config(settings.program_config_file, config_schema)
    else:
        config = None

    celery_config = load_config(settings.celery_config_file, CeleryConfig)
    app = Celery()
    app.config_from_object(celery_config.model_dump())
    return program_class(config, app)


def load_config_schema(settings: Settings) -> type[BaseModel]:
    """Load the configuration schema for the program."""
    program_name = settings.program_name_file.read_text().strip()
    program_class = load_program_class(program_name)
    return program_class.get_config_schema()


def load_worker_config(settings: Settings) -> WorkerConfig:
    """Load the configuration schema for the program."""
    program_name = settings.program_name_file.read_text().strip()
    program_class = load_program_class(program_name)
    return program_class.worker_config or DEFAULT_WORKER_CONFIG


def write_program_file(
    program_name: str,
    settings: Settings,
) -> None:
    """Write the python script with the celery app to run."""
    path = settings.program_file

    if not path.parent.exists():
        path.parent.mkdir(parents=True)

    path.write_text(
        render_template(
            "app.py.jinja2",
            program_name=program_name,
            settings=settings,
        )
    )


class ProgramState(str, Enum):
    OK = "ok"
    UNHEALTHY = "unhealthy"
    ERROR = "error"


def get_program_state(settings: Settings) -> ProgramState:
    program = load_program(settings)
    try:
        program.check(program.config)
        return ProgramState.OK
    except exceptions.HealthCheckError:
        return ProgramState.UNHEALTHY
    except:  # noqa: E722
        return ProgramState.ERROR



"""Implementation of AudioRecorder for acoupi.

An AudioRecorder is used to record audio files.
The PyAudioRecorder is implemented as class that inherit from
`AudioRecorder`. The class should implement the record() method which
return a temporary audio file based on the dataclass Recording.

The audio recorder takes argument related to the audio device. It
specifies the acoutics parameters of recording an audio file. These are
the samplerate, the duration, the number of audio_channels, the chunk
size, and the index of the audio device. The index of the audio device
corresponds to the index of the USB port the device is connected to. The
audio recorder return a temporary .wav file.
"""

import argparse
import datetime
import math
import wave
from pathlib import Path
from typing import List, Optional

import click
import pyaudio
from celery.utils.log import get_task_logger
from pydantic import BaseModel, Field

from acoupi import data
from acoupi.components.types import AudioRecorder
from acoupi.devices.audio import get_input_device_by_name, get_input_devices
from acoupi.system.config.parsers import parse_field_from_args
from acoupi.system.exceptions import HealthCheckError, ParameterError

TMP_PATH = Path("/run/shm/")

__all__ = [
    "PyAudioRecorder",
    "MicrophoneConfig",
]


class PyAudioRecorder(AudioRecorder):
    """Component that records fixed duration audio to a file."""

    duration: float
    """The duration of the audio file in seconds."""

    samplerate: int
    """The samplerate of the audio file in Hz."""

    audio_channels: int
    """The number of audio channels."""

    device_name: str
    """The name of the input audio device."""

    chunksize: int

    audio_dir: Path
    """The directory where to store the created recordings."""

    time_expansion: float
    """The time expansion factor for the recording."""

    def __init__(
        self,
        duration: float,
        samplerate: int,
        audio_channels: int,
        device_name: str,
        chunksize: int = 2048,
        audio_dir: Path = TMP_PATH,
        logger=None,
        time_expansion: float = 1,
    ) -> None:
        """Initialise the AudioRecorder with the audio parameters."""
        # Audio Duration
        self.duration = duration

        # Audio Microphone Parameters
        self.samplerate = samplerate
        self.audio_channels = audio_channels
        self.device_name = device_name

        # Audio Files Parameters
        self.chunksize = chunksize
        self.audio_dir = audio_dir
        self.sample_width = pyaudio.get_sample_size(pyaudio.paInt16)
        self.time_expansion = time_expansion

        if self.time_expansion <= 0:
            raise ValueError("time_expansion must be greater than 0")

        if logger is None:
            logger = get_task_logger(__name__)
        self.logger = logger

    def record(self, deployment: data.Deployment) -> data.Recording:
        """Record an audio file.

        Returns
        -------
        data.Recording: A Recording object containing the temporary path of the file.
        """
        now = datetime.datetime.now()
        temp_path = self.audio_dir / f"{now.strftime('%Y%m%d_%H%M%S')}.wav"
        frames = self.get_recording_data(duration=self.duration)
        self.save_recording(frames, temp_path)

        return data.Recording(
            path=temp_path,
            created_on=now,
            duration=self.duration,
            samplerate=self.samplerate,
            audio_channels=self.audio_channels,
            time_expansion=self.time_expansion,
            chunksize=self.chunksize,
            deployment=deployment,
        )

    def get_recording_data(
        self,
        duration: Optional[float] = None,
        num_chunks: Optional[int] = None,
    ) -> bytes:
        if num_chunks is None:
            if duration is None:
                raise ValueError("duration or num_chunks must be provided")

            # NOTE: Round up to the nearest chunk otherwise the recording will
            # be shorter than the requested duration
            num_chunks = math.ceil(duration * self.samplerate / self.chunksize)

        if duration is None:
            duration = num_chunks * self.chunksize / self.samplerate

        p = pyaudio.PyAudio()

        device = self.get_input_device(p)

        stream = p.open(
            format=pyaudio.paInt16,
            channels=self.audio_channels,
            rate=self.samplerate,
            input=True,
            frames_per_buffer=self.chunksize,
            input_device_index=device.index,
        )

        frames = []
        for _ in range(0, num_chunks if num_chunks > 0 else 1):
            audio_data = stream.read(
                self.chunksize,
                exception_on_overflow=True,
            )
            frames.append(audio_data)

        stream.stop_stream()
        stream.close()
        p.terminate()

        data = b"".join(frames)

        # NOTE: Due to the rounding up of the number of chunks, the recording
        # might be longer than the requested duration. We need to truncate the
        # data to the expected length.
        expected_length = int(
            2 * self.samplerate * duration * self.audio_channels
        )
        if len(data) > expected_length:
            data = data[:expected_length]

        return data

    def get_expanded_samplerate(self):
        return int(self.samplerate / self.time_expansion)

    def save_recording(self, data: bytes, path: Path) -> None:
        """Save the recording to a file."""
        with wave.open(str(path), "wb") as temp_audio_file:
            temp_audio_file.setnchannels(self.audio_channels)
            temp_audio_file.setsampwidth(self.sample_width)
            temp_audio_file.setframerate(self.get_expanded_samplerate())
            temp_audio_file.writeframes(data)

    def get_input_device(self, p: pyaudio.PyAudio):
        """Get the input device."""
        try:
            device = get_input_device_by_name(p, self.device_name)
        except IOError as error:
            raise ParameterError(
                value="device_name",
                message=(
                    "The selected input device was not found. "
                    f"Device name: {self.device_name}"
                ),
                help="Check if the microphone is connected.",
            ) from error

        return device

    def check(self):
        """Check if the audio recorder is compatible with the config."""
        num_chunks = 20
        try:
            data = self.get_recording_data(num_chunks=num_chunks)
        except ParameterError as error:
            raise HealthCheckError(
                message=(
                    "The audio recorder is not compatible with the "
                    "selected microphone. Check the configurations."
                    f"Error: {error}"
                )
            ) from error
        except OSError as error:
            if "Invalid sample rate" in str(error):
                raise HealthCheckError(
                    message=(
                        "The audio recorder is not compatible with the "
                        "selected samplerate. Check the configurations."
                    )
                ) from error

            if "Invalid number of channels" in str(error):
                raise HealthCheckError(
                    message=(
                        "The audio recorder is not compatible with the "
                        "selected number of channels. Check the configurations."
                    )
                ) from error

            raise error

        if len(data) != self.chunksize * self.audio_channels * 2 * num_chunks:
            raise HealthCheckError(
                message=(
                    "The audio recorder is not working properly. "
                    "Check the configurations."
                )
            )


class MicrophoneConfig(BaseModel):
    device_name: str
    samplerate: int = 48_000
    audio_channels: int = 1
    time_expansion: float = Field(default=1, gt=0)

    @classmethod
    def setup(
        cls,
        args: List[str],
        prompt: bool = True,
        prefix: str = "",
    ) -> "MicrophoneConfig":
        """Set up the microphone configuration."""
        return parse_microphone_config(args, prompt, prefix)


def parse_microphone_config(
    args: List[str],
    prompt: bool = True,
    prefix: str = "",
) -> MicrophoneConfig:
    # TODO: Adapt to use args, prompt and prefix
    click.secho("\n* Setting up microphone configuration.\n", fg="green")

    p = pyaudio.PyAudio()

    available_devices = get_input_devices(p)

    if len(available_devices) == 0:
        click.secho(
            "\n* No compatible audio device found.\n", fg="red", bold=True
        )
        raise ParameterError(
            value="device_index",
            message="No compatible audio device found.",
            help="Check if the microphone is connected.",
        )

    if not prompt:
        try:
            device_name = parse_field_from_args(
                "device_name",
                MicrophoneConfig.model_fields["device_name"],
                args,
                prompt=False,
                prefix=prefix,
            )

        except argparse.ArgumentError:
            device_name = available_devices[0].name

        try:
            device = next(
                d for d in available_devices if d.name == device_name
            )
        except StopIteration as error:
            raise ParameterError(
                value="device_name",
                message=f"No audio device with the name {device_name} found.",
                help="Check if the microphone is connected or review the "
                "available audio devices.",
            ) from error

        samplerate = parse_field_from_args(
            "samplerate",
            MicrophoneConfig.model_fields["samplerate"],
            args,
            prompt=False,
            prefix=prefix,
        )

        if not samplerate:
            samplerate = int(device.default_samplerate)

        channels = parse_field_from_args(
            "audio_channels",
            MicrophoneConfig.model_fields["audio_channels"],
            args,
            prompt=False,
            prefix=prefix,
        )

        time_expansion = parse_field_from_args(
            "time_expansion",
            MicrophoneConfig.model_fields["time_expansion"],
            args,
            prompt=False,
            prefix=prefix,
        )

        return MicrophoneConfig(
            device_name=device.name,
            samplerate=int(samplerate),  # type: ignore
            audio_channels=channels or 1,  # type: ignore
            time_expansion=time_expansion,  # type: ignore
        )

    click.secho("Available audio devices:\n", fg="green", bold=True)
    click.secho(
        f"{'Index':<7}{'Name':40}{'Channels':<10}{'Sample Rate':<10}",
        fg="green",
    )
    for device in available_devices:
        click.secho(f"[{device.index:>2}]   ", fg="green", bold=True, nl=False)
        click.echo(
            f"{device.name[:38]:<40}"
            f"{device.max_input_channels:<10}"
            f"{device.default_samplerate:<10}",
        )

    while True:
        try:
            index = click.prompt(
                "\nSelect an audio device index",
                type=click.Choice([str(d.index) for d in available_devices]),
                value_proc=int,
            )
            selected_device = next(
                d for d in available_devices if d.index == index
            )
            break
        except (ValueError, StopIteration):
            click.secho(
                "Invalid input. Please select a valid audio device index.",
                fg="red",
            )

    click.secho("\nInfo of selected audio device:\n", fg="green", bold=True)
    click.secho(
        f"{'index':20} = {selected_device.index}\n"
        f"{'name':20} = {selected_device.name}\n"
        f"{'max channels':20} = {selected_device.max_input_channels}\n"
        f"{'default sample rate':20} = {selected_device.default_samplerate}\n",
        fg="yellow",
    )

    while True:
        try:
            channels = click.prompt(
                "Select the number of audio channels",
                type=click.Choice(
                    [
                        str(i)
                        for i in range(
                            1, 1 + selected_device.max_input_channels
                        )
                    ]
                ),
                value_proc=int,
                default=selected_device.max_input_channels,
            )
            if channels > selected_device.max_input_channels:
                raise ValueError
            break
        except ValueError:
            click.secho(
                "Invalid input. Please select a valid "
                "number of audio channels.",
                fg="red",
            )

    while True:
        try:
            samplerate = click.prompt(
                "\nSelect the samplerate. The default samplerate is "
                "recommended but your device might support other "
                "sampling rates.",
                type=int,
                default=selected_device.default_samplerate,
            )
            if p.is_format_supported(
                samplerate,
                input_device=selected_device.index,
                input_channels=channels,
                input_format=pyaudio.paInt16,
            ):
                break
        except ValueError as error:
            if error.args[0] == "Invalid sample rate":
                click.secho(
                    "Samplerate not supported by the audio device.",
                    fg="red",
                )
                click.secho(
                    "Please try with another samplerate. "
                    "We recommend you consult the documentation of your "
                    "audio device to find the supported samplerates.",
                    fg="yellow",
                )
            else:
                click.secho(
                    "Invalid input. Please select a valid samplerate.",
                    fg="red",
                )

    time_expansion = click.prompt(
        "Adjust the playback speed/duration metadata for downstream analysis.\n"
        "Example: Enter 10.0 if you recorded at 480kHz but need to process at 48kHz.\n"
        "Enter the time expansion factor (default: 1.0). \n",
        type=click.FloatRange(min=0.0, clamp=False, min_open=True),
        default=1.0,
    )

    return MicrophoneConfig(
        device_name=selected_device.name,
        samplerate=samplerate,
        audio_channels=channels,
        time_expansion=time_expansion,
    )



from pathlib import Path

import pytest

from acoupi.components.audio_recorder import MicrophoneConfig
from acoupi.components.messengers import HTTPConfig
from acoupi.programs.templates import (
    AudioConfiguration,
    MessagingConfig,
    PathsConfiguration,
)
from acoupi.system.constants import CeleryConfig


@pytest.fixture(scope="session")
def celery_worker_parameters():
    """Redefine this fixture to change the init parameters of Celery workers.

    This can be used e. g. to define queues the worker will consume tasks from.

    The dict returned by your fixture will then be used
    as parameters when instantiating :class:`~celery.worker.WorkController`.
    """
    return {
        "loglevel": "WARN",
        "queues": ("celery", "default", "special"),
    }


@pytest.fixture
def paths_config(tmp_path: Path) -> PathsConfiguration:
    return PathsConfiguration(
        tmp_audio=tmp_path / "tmp",
        recordings=tmp_path / "audio",
        db_metadata=tmp_path / "metadata.db",
    )


@pytest.fixture
def audio_config() -> AudioConfiguration:
    return AudioConfiguration(duration=1, interval=2)


@pytest.fixture
def microphone_config():
    return MicrophoneConfig(
        samplerate=44100,
        audio_channels=1,
        device_name="default",
    )


@pytest.fixture
def messaging_config(tmp_path: Path) -> MessagingConfig:
    return MessagingConfig(
        messages_db=tmp_path / "messages.db",
        http=HTTPConfig(
            base_url="http://localhost:8000",
        ),
    )



"""Saving managers for the recordings and detections of acoupi.

Saving managers are used to determine where and how the recordings and
detections of an audio file should be saved. This is helpful to handle
recordings files and detections outputs. Recordings and detections outputs
can be saved into a specific format (i.e, .wav files, .csv files) and at a
specific location (i.e, rpi memory, external hardrive, folder XX/YY).

The SavingManagers are implemented as class that inherit from RecordingSavingManager.
The classes should implement the save_recording method. The save_recording method
takes a recording object and a list of model outputs as input and returns the path where
the recording should be saved.

The save_recording method is called by the file management task to determine where the
recording should be saved. The file management task is responsible for moving recordings
from the memory to the disk, and remove recordings that are no longer needed.

The SavingManagers are optional and can be ingored if the no recordings are saved.
"""

import logging
import os
import string
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional

from celery.utils.log import get_task_logger

from acoupi import data
from acoupi.components import types
from acoupi.devices import get_device_info

__all__ = [
    "SaveRecordingManager",
    "IDFileManager",
    "DateFileManager",
]


class SaveRecordingManager(types.RecordingSavingManager):
    """A Recording SavingManager that save audio recordings."""

    dirpath: Path
    """Directory path to save recordings."""

    dirpath_true: Path
    """Directory path to save recordings if audio recording contains confident detections
    (i.e., above the detection threshold)."""

    dirpath_false: Path
    """Directory path to save recordings if audio recording contain no confident detections
    (i.e., below the detection threshold)."""

    timeformat: str
    """Datetime format to use to name the recording file path."""

    detection_threshold: float
    """Threshold determining if a recording contains confident detections."""

    saving_threshold: float
    """Threshold determining if recordings should be saved (i.e., regardless of confident
    or unconfident detections)."""

    def __init__(
        self,
        dirpath: Path,
        dirpath_true: Optional[Path] = None,
        dirpath_false: Optional[Path] = None,
        timeformat: str = "%Y%m%d_%H%M%S",
        detection_threshold: float = 0.6,
        saving_threshold: float = 0.3,
        logger: Optional[logging.Logger] = None,
    ):
        """Initiatilise the Recording SavingManager."""
        if not dirpath.exists():
            dirpath.mkdir(parents=True)
        self.dirpath = dirpath

        if dirpath_true is None:
            dirpath_true = dirpath / "true_detections"
        if not dirpath_true.exists():
            dirpath_true.mkdir(parents=True)
        self.dirpath_true = dirpath_true

        if dirpath_false is None:
            dirpath_false = dirpath / "false_detections"
        if not dirpath_false.exists():
            dirpath_false.mkdir(parents=True)
        self.dirpath_false = dirpath_false

        self.timeformat = timeformat
        self.detection_threshold = detection_threshold
        self.saving_threshold = saving_threshold

        if logger is None:
            logger = get_task_logger(__name__)
        self.logger = logger

    def get_saving_recording_path(
        self,
        model_outputs: Optional[List[data.ModelOutput]],
    ) -> Optional[Path]:
        """Determine where the recording should be saved.

        Parameters
        ----------
        model_outputs : Optional[List[data.ModelOutput]]
            List of model outputs containing detections and tags.

        Returns
        -------
        Path
            Path where the recording should be saved.
        """
        if not model_outputs:
            return self.dirpath

        for model_output in model_outputs:
            # Check if any tags or detectinos are confident
            if any(
                tag.confidence_score >= self.detection_threshold
                for tag in model_output.tags
            ):
                return self.dirpath_true

            if any(
                detection.detection_score >= self.detection_threshold
                for detection in model_output.detections
            ):
                return self.dirpath_true

            if any(
                tag.confidence_score >= self.saving_threshold
                for tag in model_output.tags
            ):
                return self.dirpath_false

            if any(
                detection.detection_score >= self.saving_threshold
                for detection in model_output.detections
            ):
                return self.dirpath_false

        return None

    def save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> Optional[Path]:
        """Save a recording to a file.

        Examples
        --------
        >>> dirpath = Path("path/to/save")
        >>> dirpath_true = Path("path/to/save/confident_detections")
        >>> dirpath_false = Path("path/to/save/unconfident_detections")
        >>> detection_threshold = 0.8
        >>> saving_threshold = 0.3

        >>> model_outputs = [
        ...     data.ModelOutput(tags=[data.Tag(confidence_score=0.7)]),
        ...     data.ModelOutput(
        ...         detections=[data.Detection(detection_score=0.6)]
        ...     ),
        ... ]
        >>> saving_directory = self.get_saving_recording_path(
        ...     model_outputs
        ... )
        >>> assert saving_directory == dirpath_false
        """
        if recording.path is None:
            raise ValueError("Recording has no path")

        sdir = self.get_saving_recording_path(model_outputs)

        if sdir is None:
            return None

        srec_filename = recording.created_on.strftime(self.timeformat)

        # Move recording to the path it should be saved
        new_path = sdir / f"{srec_filename}.wav"
        return new_path


class BaseFileManager(types.RecordingSavingManager, ABC):
    """Base implementation for simple recording saving managers.

    This class can be used to implement simple recording saving managers
    that do not use model outputs to determine where the recording
    should be saved.

    All recordings are saved in a directory specified in the constructor
    and the relative path is determined by the get_file_path method.
    """

    directory: Path
    """Directory where the files are stored."""

    def __init__(
        self, directory: Path, logger: Optional[logging.Logger] = None
    ):
        """Create a new BaseFileManager."""
        if logger is None:
            logger = get_task_logger(__name__)
        self.logger = logger
        self.directory = directory

        if not os.path.exists(directory):
            os.makedirs(directory)

    @abstractmethod
    def get_file_path(self, recording: data.Recording) -> Path:
        """Get the path where the file of a recording should be stored.

        The path must be relative to the directory specified in the
        constructor.

        Parameters
        ----------
        recording: data.Recording
            Recording to get the path for.

        Returns
        -------
            Path of the file.
        """
        raise NotImplementedError

    def save_recording(
        self,
        recording: data.Recording,
        model_outputs: Optional[List[data.ModelOutput]] = None,
    ) -> Path:
        """Save a recording to a file.

        Parameters
        ----------
        recording: data.Recording
            Recording to save.

        Returns
        -------
            Path of the saved file.
        """
        if not recording.path:
            raise ValueError("Recording has no path")

        if not recording.path.is_file():
            raise FileNotFoundError("Recording file does not exist")

        # Get the path where the file should be stored
        path = self.get_file_path(recording)

        # Add the base directory as prefix
        full_path = self.directory / path

        # Get the directory where the file should be stored
        directory = full_path.parent

        # Create the directory if it does not exist
        if not os.path.exists(directory):
            os.makedirs(directory)

        return full_path


_SANITIZE_TABLE = str.maketrans({char: "_" for char in '<>:"/\\|?*'})


def sanitize_filename(filename: str) -> str:
    """Replace characters that are unsafe in filenames."""
    return filename.translate(_SANITIZE_TABLE)


class FilenameFormatter(string.Formatter):
    """Format filename templates while treating ``None`` as an empty string.

    This keeps optional context fields, such as deployment coordinates or the
    device id, from raising formatting errors when they are missing.
    """

    def get_value(self, key, args, kwargs):
        value = super().get_value(key, args, kwargs)
        if value is None:
            return ""
        return value

    def format_field(self, value, format_spec):
        if value is None:
            value = ""
        return super().format_field(value, format_spec)


class DateFileManager(BaseFileManager):
    """Save recordings into date-based folders.

    Recordings are stored in a directory structure based on the recording
    date:

    ``YYYY/MM/DD/``

    By default, each filename includes the recording date, time, and the
    recording ID:

    ``YYYYMMDD_HHMMSS_<recording-id>.wav``

    Examples
    --------
    Use the default filename pattern:

    >>> manager = DateFileManager(Path("recordings"))
    >>> recording = data.Recording(
    ...     path=Path("tmp.wav"),
    ...     duration=1.0,
    ...     samplerate=48000,
    ...     deployment=data.Deployment(name="site-a"),
    ... )
    >>> relative_path = manager.get_file_path(recording)
    >>> relative_path.parts[:3] == (
    ...     str(recording.created_on.year),
    ...     str(recording.created_on.month),
    ...     str(recording.created_on.day),
    ... )
    True

    Notes
    -----
    The filename can be customized by providing a different
    ``filename_template``. The template uses Python's standard format-string
    syntax and can reference ``recording``, ``deployment``, and ``device``.
    Use ``recording.created_on`` with datetime format specifiers when you need
    year, month, day, or time components in the filename.

    Build a more descriptive filename from deployment and recording fields:

    >>> manager = DateFileManager(
    ...     Path("recordings"),
    ...     filename_template=(
    ...         "{deployment.name}_{recording.created_on:%Y%m%d_%H%M%S}"
    ...         "_{recording.samplerate}Hz.wav"
    ...     ),
    ... )

    Include the runtime device id when available:

    >>> manager = DateFileManager(
    ...     Path("recordings"),
    ...     filename_template=(
    ...         "{deployment.name}_{device.id}_"
    ...         "{recording.created_on:%H%M%S}.wav"
    ...     ),
    ... )

    If you need a different directory layout altogether, use another file
    manager or subclass ``BaseFileManager``.
    """

    filename_template: str

    def __init__(
        self,
        directory: Path,
        filename_template: str = (
            "{recording.created_on:%Y%m%d_%H%M%S}_{recording.id}.wav"
        ),
        logger: Optional[logging.Logger] = None,
    ):
        """Create a date-based file manager.

        Parameters
        ----------
        directory:
            Root directory where recordings are stored.
        filename_template:
            Format string used to build the filename within the date-based
            directory. The template can reference ``recording``, ``deployment``
            and ``device`` fields, for example
            ``"{deployment.name}_{recording.created_on:%Y%m%d_%H%M%S}.wav"``.
        logger:
            Optional logger used by the manager.
        """
        super().__init__(directory=directory, logger=logger)
        self.filename_template = filename_template

    def _render_filename(self, recording: data.Recording) -> str:
        """Render and sanitize the filename for a recording.

        Raises
        ------
        ValueError
            If the template references an unknown field or contains an invalid
            format specification.
        """
        formatter = FilenameFormatter()

        device = get_device_info()

        try:
            filename = formatter.format(
                self.filename_template,
                recording=recording,
                deployment=recording.deployment,
                device=device,
            )
        except (AttributeError, KeyError, ValueError) as error:
            raise ValueError(
                f"Invalid filename template: {self.filename_template}"
            ) from error

        return sanitize_filename(filename)

    def get_file_path(self, recording: data.Recording) -> Path:
        """Get the path where the file of a recording should be stored.

        Parameters
        ----------
        recording
            Recording to get the path for.

        Returns
        -------
            Path of the file.

        Notes
        -----
        The returned path is relative to ``self.directory`` and always follows
        the ``YYYY/MM/DD/<filename>`` structure.
        """
        date = recording.created_on
        directory = (
            Path(str(date.year)) / Path(str(date.month)) / Path(str(date.day))
        )
        return directory / Path(self._render_filename(recording))


class IDFileManager(BaseFileManager):
    """FileManager that uses the ID of the recording to organise the files.

    The recordings are saved in a single directory that is specified in
    the constructor. The files are named using the ID of the recording.
    The format is

    ID.wav
    """

    def get_file_path(self, recording: data.Recording) -> Path:
        """Get the the path where the file of a recording should be stored."""
        return Path(f"{recording.id}.wav")



"""Custom Exceptions for Acoupi system."""

from typing import Optional

__all__ = [
    "ConfigurationError",
    "DeploymentError",
    "HealthCheckError",
    "InvalidProgramError",
    "MetadataStoreError",
    "MessageSendError",
    "MessageStoreError",
    "ParameterError",
    "ProgramNotFoundError",
]


class ProgramNotFoundError(Exception):
    """Exception raised when program is not found."""

    def __init__(self, program: str):
        """Initialise ProgramNotFoundError exception."""
        self.program = program
        super().__init__(program)


class InvalidProgramError(Exception):
    """Exception raised when a program file is invalid."""

    def __init__(self, program: str):
        """Initialise InvalidProgramError exception."""
        self.program = program
        super().__init__(program)


class ConfigurationError(Exception):
    """Exception raised when a configuration is invalid."""

    def __init__(self, message: str, help: Optional[str] = None):
        """Initialise ConfigurationError exception.

        Parameters
        ----------
        message : str
            The error message.
        help : str, optional
            An optional help message on how to fix the error.
        """
        self.message = message
        self.help = help
        super().__init__(message, help)

    def __str__(self):
        """Return the error message."""
        return self.message


class ParameterError(Exception):
    """Exception raised when a parameter is invalid."""

    def __init__(self, value: str, message: str, help: Optional[str] = None):
        """Initialise ParameterError exception.

        Parameters
        ----------
        value : str
            The value that caused the error.
        message : str
            The error message.
        help : str, optional
            An optional help message on how to fix the error.
        """
        self.value = value
        self.message = message
        self.help = help
        super().__init__(value, message, help)

    def __str__(self):
        """Return the error message."""
        return self.message


class HealthCheckError(Exception):
    """Exception raised when a health check fails."""

    def __init__(self, message: str):
        """Initialise HealthError exception."""
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Return the error message."""
        return self.message


class DeploymentError(Exception):
    """Exception raised when a deployment fails."""

    def __init__(self, message: str):
        """Initialise DeploymentError exception."""
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Return the error message."""
        return self.message


class MessageSendError(Exception):
    """Exception raised when a message cannot be sent locally."""

    def __init__(self, message: str):
        """Initialise MessageSendError exception."""
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Return the error message."""
        return self.message


class MessageStoreError(Exception):
    """Exception raised when the message store state is invalid."""

    def __init__(self, message: str):
        """Initialise MessageStoreError exception."""
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Return the error message."""
        return self.message


class MetadataStoreError(Exception):
    """Exception raised when the metadata store state is invalid."""

    def __init__(self, message: str):
        """Initialise MetadataStoreError exception."""
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Return the error message."""
        return self.message



"""Message store Sqlite implementation."""

from pathlib import Path
from typing import List, Union

from pony import orm

from . import types as db_types
from .database import create_message_models
from acoupi import data
from acoupi.components import types
from acoupi.system.exceptions import MessageStoreError


class SqliteMessageStore(types.MessageStore):
    """Message store Sqlite implementation.

    This store keeps track of messages that have been sent to a remote server.
    It stores the response of the server and the datetime when the message was
    sent. This allows the client to keep track of which messages have been
    sent and which have not, allowing it to send them again if necessary.

    The messages are stored in a local sqlite database.
    """

    db_path: Path
    """Path to the database file."""

    database: orm.Database
    """The Pony ORM database object."""

    models: db_types.MessageModels
    """The Pony ORM models."""

    def __init__(self, db_path: Path) -> None:
        """Initialise the message store."""
        self.db_path = db_path
        self.database = orm.Database()
        self.models = create_message_models(self.database)
        self.database.bind(
            provider="sqlite",
            filename=str(db_path),
            create_db=True,
        )
        self.database.generate_mapping(create_tables=True)

    @orm.db_session
    def store_message(
        self,
        message: data.Message,
    ) -> None:
        """Register a recording message with the store."""
        return self._create_message(message)

    @orm.db_session
    def get_unsent_messages(self) -> List[data.Message]:
        """Get the messages that have not been synced to the server."""
        # Get messages that haven't had an OK response
        unsent_messages = self.models.Message.select(
            lambda m: not orm.exists(r for r in m.responses if r.status == 0)
        )

        return [
            data.Message(
                id=message.id,
                content=message.content,
                created_on=message.created_on,
            )
            for message in unsent_messages
        ]

    @orm.db_session
    def store_response(
        self,
        response: data.Response,
    ) -> None:
        """Store a response to a message."""
        try:
            db_message = self.models.Message[response.message.id]
        except orm.ObjectNotFound as error:
            raise MessageStoreError(
                f"Cannot store response for unknown message {response.message.id}."
            ) from error

        self.models.Response(
            message=db_message,
            content=response.content or "",
            status=response.status.value,
            received_on=response.received_on,
        )
        orm.commit()

    @orm.db_session
    def _create_message(
        self,
        message: data.Message,
    ) -> db_types.Message:
        """Create a Pony ORM message object."""
        db_message = self.models.Message(
            id=message.id,
            content=self._as_bytes(message.content),
            created_on=message.created_on,
        )
        orm.commit()
        return db_message

    @staticmethod
    def _as_bytes(content: Union[str, bytes]) -> bytes:
        """Normalise message content to bytes for storage."""
        if isinstance(content, bytes):
            return content

        return content.encode("utf-8")



"""Test output cleaners."""

import datetime
from pathlib import Path
from typing import List

import pytest

from acoupi import data
from acoupi.components import output_cleaners


@pytest.fixture
def create_test_model_output():
    """Return a model output."""
    deployment = data.Deployment(
        name="test_deployment",
    )

    recording = data.Recording(
        path=Path("test.wav"),
        duration=1,
        samplerate=256000,
        audio_channels=1,
        deployment=deployment,
        created_on=datetime.datetime.now(),
    )

    def factory(
        detections: List[data.Detection],
    ) -> data.ModelOutput:
        """Return a model output."""
        return data.ModelOutput(
            recording=recording,
            name_model="test_model",
            detections=detections,
        )

    return factory


@pytest.fixture
def create_test_detection():
    """Fixture for creating random detections.

    Will create detections with no location and a single tag.
    """

    def factory(
        tag_value: str,
        detection_score: float = 0.5,
        tag_score: float = 1,
        tag_key: str = "species",
    ) -> data.Detection:
        """Return a random detection."""
        return data.Detection(
            detection_score=detection_score,
            tags=[
                data.PredictedTag(
                    tag=data.Tag(
                        value=tag_value,
                        key=tag_key,
                    ),
                    confidence_score=tag_score,
                ),
            ],
        )

    return factory


def test_threshold_detection_filter_removes_detections_with_low_confidence(
    create_test_model_output,
    create_test_detection,
):
    """Test threshold filter removes detections with low confidence."""
    # Arrange
    cleaner = output_cleaners.ThresholdDetectionCleaner(
        detection_threshold=0.5,
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                "species1",
                detection_score=0.1,
                tag_score=0.5,
            ),
            create_test_detection(
                "species2",
                detection_score=0.6,
                tag_score=0.5,
            ),
        ]
    )

    # Act
    cleaned_model_output = cleaner.clean(model_output)

    # Assert
    assert len(cleaned_model_output.detections) == 1
    assert cleaned_model_output.detections[0].tags[0].tag.value == "species2"


def test_threshold_detection_keeps_detections_even_with_low_confidence_tags(
    create_test_model_output,
    create_test_detection,
):
    """Test threshold filter keeps detections even with low confidence tags."""
    cleaner = output_cleaners.ThresholdDetectionCleaner(
        detection_threshold=0.5,
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                "species1",
                detection_score=0.7,
                tag_score=0.1,
            ),
            create_test_detection(
                "species2",
                detection_score=0.6,
                tag_score=0.5,
            ),
        ]
    )

    # Act
    cleaned_model_output = cleaner.clean(model_output)

    # Assert
    assert len(cleaned_model_output.detections) == 2


def test_threshold_removes_detections_with_default_tag_score(
    create_test_model_output,
    create_test_detection,
):
    """Test threshold filter removes detections with default tag score."""
    # Arrange
    cleaner = output_cleaners.ThresholdDetectionCleaner(
        detection_threshold=0.5,
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                "species1",
                detection_score=0.7,
                tag_score=0.5,
            ),
            create_test_detection(
                "species2",
                detection_score=0.3,
            ),
        ]
    )

    # Act
    cleaned_model_output = cleaner.clean(model_output)

    # Assert

    assert len(cleaned_model_output.detections) == 1
    assert cleaned_model_output.detections[0].tags[0].tag.value == "species1"


def test_threshold_removes_low_score_tags(
    create_test_model_output,
    create_test_detection,
):
    """Test filter keeps tags even if with low score score."""
    # Arrange
    cleaner = output_cleaners.ThresholdDetectionCleaner(
        detection_threshold=0.5,
    )
    model_output = create_test_model_output(
        detections=[
            create_test_detection(
                "species1",
                detection_score=0.7,
                tag_score=0.1,
            ),
        ]
    )

    # Act
    cleaned_model_output = cleaner.clean(model_output)

    # Assert
    assert len(cleaned_model_output.detections) == 1
    assert len(cleaned_model_output.detections[0].tags) == 0



"""Acoupi devices module.

This module provides functions to retrieve information about the device
on which Acoupi is running. This includes identifying the device type
(currently supports Raspberry Pi), and obtaining unique identifiers
like serial numbers and hostnames.
"""

import uuid
from functools import cache

from acoupi.data import DeviceInfo
from acoupi.devices.rpi import get_rpi_host_name, get_rpi_serial_number, is_rpi


def get_device_id() -> str:
    """Get a unique identifier for the current device.

    This function returns a device-specific identifier.

    - If the device is a Raspberry Pi, the serial number is returned.
    - Otherwise, the MAC address is used as the identifier,
      retrieved via `uuid.getnode()`.

    Returns
    -------
    str
        The device ID.

    Notes
    -----
        For potential limitations of using MAC addresses for identification,
        refer to the [`uuid.getnode()`][uuid.getnode] documentation.
    """
    if is_rpi():
        return get_rpi_serial_number()

    return str(uuid.getnode())


@cache
def get_device_info() -> DeviceInfo:
    """Get cached runtime information about the current device.

    Returns
    -------
    DeviceInfo
        Device information for the current process. If the device identifier
        cannot be resolved, the returned object contains an empty ``id``.
    """
    try:
        return DeviceInfo(id=get_device_id())
    except Exception:
        return DeviceInfo(id="")


__all__ = [
    "get_rpi_serial_number",
    "get_rpi_host_name",
    "get_device_id",
    "get_device_info",
    "is_rpi",
]
